! $Id: BlockConstructs.f90 2193 2016-06-06 13:01:31Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BlockConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /block-construct/'s.

MODULE BlockConstructs
  
  USE Identifiers
  USE Labels
  USE ParseNodes
  USE SourceLocations
  USE SpecificationParts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseBlockConstruct
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /block-construct/.
  !!
  !! We could have extended BlockStorage for this, but then you potentially 
  !! have confusion about whether the children are executable children or 
  !! specification children.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: BlockConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The results of parsing the specification part.
    !!
    !! Not allocated if the block had no specification part.
    TYPE(SpecificationPart), ALLOCATABLE :: spec
    
    !> The results of parsing the executable children.
    !!
    !! Always allocated, may be size zero.
    TYPE(SCParseNodeElement), ALLOCATABLE :: exec_children(:)
    !> The label on the end block statement, if present, not allocated 
    !! otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => block_Dump
    PROCEDURE, NOPASS :: What => block_What
    PROCEDURE :: Query => block_Query
    PROCEDURE :: Modify => block_Modify
    PROCEDURE :: Advertise => block_Advertise
  END TYPE BlockConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for a /block-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    block_construct   The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseBlockConstruct(st, block_construct, err_list)
    
    USE BlockConstructStNodes
    USE BlockStmts
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(BlockConstructStNode), INTENT(INOUT) :: st
    TYPE(BlockConstruct), INTENT(OUT) :: block_construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening block statement.
    
    CALL ParseBlockStmt( st%first%tlist, block_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Specification part.
    
    IF (ALLOCATED(st%spec)) THEN
      ALLOCATE(block_construct%spec)
      CALL ParseSpecificationPart(st%spec, block_construct%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    !---------------------------------------------------------------------------
    ! Executable children.
    
    IF (ALLOCATED(st%exec)) THEN
      CALL ParseBlock(st%exec, block_construct%exec_children, sub_err_list)
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(block_construct%exec_children(0))
    END IF
    
    !---------------------------------------------------------------------------
    ! The closing end block statement.
    
    CALL ParseEndBlockStmt(  &
        st%last%tlist,  &
        block_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      block_construct%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseBlockConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockConstruct%Dump - write a debugging 
  !! representation of the results of parsing a block construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE block_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    IF (ALLOCATED(pnode%spec)) THEN
      WRITE (unit, "(A,'/specification-part/')") pfx
      CALL pnode%spec%Dump(unit, pfx // '  ')
    END IF
    
    WRITE (unit, "(A,'child executable statements:')") pfx
    CALL DumpBlock(pnode%exec_children, unit, pfx // '  ')
    
  END SUBROUTINE block_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'block-construct'.
  
  FUNCTION block_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'block-construct'
    
  END FUNCTION block_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE block_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%spec)) THEN
      ! Forward the query to the specification part.
      CALL pnode%spec%Query(visitor)
    END IF
    
    ! Forward to the child executable constructs.
    DO i = 1, SIZE(pnode%exec_children)
      CALL pnode%exec_children(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE block_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE block_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%spec)) THEN
      ! Forward the query to the specification part.
      CALL pnode%spec%Modify(visitor)
    END IF
    
    ! Forward to the child executable constructs.
    DO i = 1, SIZE(pnode%exec_children)
      CALL pnode%exec_children(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE block_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE block_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE block_Advertise
  
END MODULE BlockConstructs
