! $Id: BlockDataStmts.f90 2822 2019-03-24 07:20:23Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BlockDataStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /block-data-stmt/'s and /end-block-data-stmt/'s.

MODULE BlockDataStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: ParseBlockDataStmt
  PUBLIC :: ParseEndBlockDataStmt
  
  !> Stores the results of parsing a /block-data-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: BlockDataStmt
    !> The /block-data-name/, if present.
    !!
    !! Not allocated if no name is provided.
    TYPE(Identifier), ALLOCATABLE :: block_data_name
  CONTAINS
    PROCEDURE :: Dump => bd_Dump
    PROCEDURE :: Query => bd_Query
    PROCEDURE :: Modify => bd_Modify
    PROCEDURE, NOPASS :: What => bd_What
    PROCEDURE :: Advertise => bd_Advertise
  END TYPE BlockDataStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /block-data-stmt/ (R1121).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    bd_stmt           The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseBlockDataStmt(tlist, bd_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
  
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(BlockDataStmt), INTENT(OUT) :: bd_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /block-data-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1121'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: it             ! Token index.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /block-data-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'BLOCK', scck_'DATA', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That can be it.
    IF (it > SIZE(tlist)) RETURN
    
    CALL ExpectName(tlist, it, '/block-data-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !===========================================================================
    ! ifort bug.
    ! bd_stmt%block_data_name = IdentifierIfortBug(tlist(it))
    ALLOCATE(bd_stmt%block_data_name, SOURCE=IdentifierIfortBug(tlist(it)))
    !===========================================================================
    
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseBlockDataStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-block-data-stmt/ (R1122).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /block-data-name/ from the statement, 
  !! or not allocated if there was no name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndBlockDataStmt(tlist, name, err_list)
    
    USE CharUtils
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /end-block-data-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1122'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-block-data-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! END [ BLOCK DATA [ name ] ] or ENDBLOCKDATA [ name ]
    
    CALL ExpectKeyword( tlist, 1,  &
        [ scck_'END         ',  &
          scck_'ENDBLOCK    ',  &
          scck_'ENDBLOCKDATA' ],  &
        '"END", "ENDBLOCK" or "ENDBLOCKDATA"', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)    ! Got END
      ! BLOCKDATA (splittable) could be next, or the statement could be 
      ! an END on its own.
      CALL Expect( &
          tlist,  &
          2,  &
          [ittName, ittName, ittEndStatement],  &
          [scck_'BLOCK    ', scck_'BLOCKDATA', scck_'         '],  &
          opt, &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      SELECT CASE (opt)
      CASE (1)  ! got BLOCK
        ! Need DATA.
        CALL ExpectKeyword(tlist, 3, scck_'DATA', err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        it = 4
      CASE (2)  ! got BLOCKDATA
        it = 3
      CASE (3)  ! got end of statement.
        RETURN
      CASE DEFAULT
        STOP 'Internal error in BlockDataStmts%%ParseEndBlockDataStmt, &
            &unexpected BLOCK[DATA] keyword option.'
      END SELECT
      
    CASE (2)    ! got ENDBLOCK
      ! Need DATA.
      CALL ExpectKeyword( tlist, 2, scck_'DATA', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = 3
      
    CASE (3)    ! got ENDBLOCKDATA
      it = 2
      
    CASE DEFAULT
      STOP 'Internal error in BlockDataStmts%%ParseEndBlockDataStmt, &
          &unexpected END[BLOCK[DATA]] keyword option.'
      
    END SELECT
    
    ! it here one beyond the END BLOCK DATA keyword sequence.
    
    !---------------------------------------------------------------------------
    ! The name is optional
    
    CALL Expect(  &
        tlist,  &
        it,  &
        [ittName, ittEndStatement],  &
        [scck_'', scck_''],  &
        '/block-data-name/ or end-of-statement',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 2) RETURN
    
    ALLOCATE(name)      ! ifort bug - should be automatically allocated.
    name = IdentifierIfortBug(tlist(it))
    
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseEndBlockDataStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockDataStmt%Dump - write a debugging 
  !! representation of the results of parsing the separate module procedure 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE bd_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockDataStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%block_data_name)) THEN
      WRITE (unit, "(A,'/block-data-name/ = ',A)") pfx,  &
          .NameOf. pnode%block_data_name
    END IF
    
  END SUBROUTINE bd_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockDataStmt%Query - traverse the node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE bd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockDataStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockDataStmt%Modify - traverse the node 
  !! with a modify operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE bd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockDataStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockDataStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'block-data-stmt'
  
  FUNCTION bd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'block-data-stmt'
    
  END FUNCTION bd_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockDataStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the separate subprogram
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE bd_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockDataStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%block_data_name)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%block_data_name,  &
          NATURES=[NatureDefBlockData],  &
          LOCATIONS=[.Locationof. pnode%block_data_name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE bd_Advertise
  
END MODULE BlockDataStmts
