! $Id: CaseConstructs.f90 2193 2016-06-06 13:01:31Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CaseConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /case-construct/'s.

MODULE CaseConstructs
  
  USE Expressions
  USE Identifiers
  USE Labels
  USE ParseNodes
  USE SelectCaseStmts
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCaseConstruct
  
  !-----------------------------------------------------------------------------
  
  !> Represents the results of parsing a case statement and its associated 
  !! block.
  !!
  !! @todo Make this an extension of BlockStorage?
  TYPE :: case_stmt_block
    !> The results of parsing the values and value ranges nominated by the 
    !! case statement.  Zero sized if this is the DEFAULT block.
    TYPE(CaseValueRange), ALLOCATABLE :: values(:)
    !> The results of parsing the executable children.
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
  END TYPE case_stmt_block
  
  !> Represents the results of parsing a /case-construct/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CaseConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The /case-expr/.
    CLASS(Expression), ALLOCATABLE :: expr
    !> The case statements and following executable statements.
    TYPE(case_stmt_block), ALLOCATABLE :: blocks(:)
    !> The label on the end block statement, if present, not allocated 
    !! otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => case_Dump
    PROCEDURE, NOPASS :: What => case_What
    PROCEDURE :: Query => case_Query
    PROCEDURE :: Modify => case_Modify
    PROCEDURE :: Advertise => case_Advertise
  END TYPE CaseConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for a /case-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    construct         The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCaseConstruct(st, construct, err_list)
    
    USE CaseConstructStNodes
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(CaseConstructStNode), INTENT(INOUT) :: st
    TYPE(CaseConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! case block index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening case statement.
    
    CALL ParseSelectCaseStmt(  &
        st%first%tlist,  &
        construct%expr,  &
        construct%construct_name, sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Case blocks.
    
    ALLOCATE(construct%blocks(SIZE(st%blocks)))
    DO i = 1, SIZE(st%blocks)
      ASSOCIATE(child_st => st%blocks(i), child_pn => construct%blocks(i))
        CALL ParseCaseStmt(  &
            child_st%initial%tlist,  &
            construct%construct_name,  &
            child_pn%values,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
      
        CALL ParseBlock(child_st%exec, child_pn%children, sub_err_list)
        CALL Add(err_list, sub_err_list)
      END ASSOCIATE
    END DO
    
    !---------------------------------------------------------------------------
    ! The closing end select case statement.
    
    CALL ParseEndSelectStmt(  &
        st%last%tlist,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      construct%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseCaseConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseConstruct%Dump - write a debugging 
  !! representation of the results of parsing a case construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE case_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! case block index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    ! The case expression.
    WRITE (unit, "(A,'/case-expr/ = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, "()")
    
    DO i = 1, SIZE(pnode%blocks)
      WRITE (unit, "(A,'/case-selector/ = ')", ADVANCE='NO') pfx
      CALL WriteCaseSelector(unit, pnode%blocks(i)%values)
      WRITE (unit, "()")
      
      WRITE (unit, "(A,'child executable statements:')") pfx
      CALL DumpBlock(pnode%blocks(i)%children, unit, pfx // '  ')
    END DO
    
  END SUBROUTINE case_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'case-construct'.
  
  FUNCTION case_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'case-construct'
    
  END FUNCTION case_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE case_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(CaseConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    
    ! Child block child statement index, or case value index.
    INTEGER :: j
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    DO i = 1, SIZE(pnode%blocks)
      ! Zero trip loop if the block is the DEFAULT block.
      DO j = 1, SIZE(pnode%blocks(i)%values)
        CALL pnode%blocks(i)%values(j)%Query(visitor)
      END DO
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Query(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE case_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE case_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(CaseConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    
    ! Child block child statement index, or case value index.
    INTEGER :: j
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    DO i = 1, SIZE(pnode%blocks)
      ! Zero trip loop if the block is the DEFAULT block.
      DO j = 1, SIZE(pnode%blocks(i)%values)
        CALL pnode%blocks(i)%values(j)%Modify(visitor)
      END DO
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Modify(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE case_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE case_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE case_Advertise
  
END MODULE CaseConstructs
