! $Id: CloseStmts.f90 2825 2019-03-24 09:14:21Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CloseStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /close-stmt/'s (R908).

MODULE CloseStmts
  
  USE Expressions
  USE Labels
  USE ParseNodes
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCloseStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /close-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CloseStmt
    !> The UNIT specification.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The IOMSG specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The ERR specification.
    !!
    !! Not allocated if there was no ERR specifier.
    TYPE(Label), ALLOCATABLE :: err
    !> The STATUS specification.
    CLASS(Expression), ALLOCATABLE :: status
  CONTAINS
    PROCEDURE :: Dump => close_Dump
    PROCEDURE :: Query => close_Query
    PROCEDURE :: Modify => close_Modify
    PROCEDURE, NOPASS :: What => close_What
  END TYPE CloseStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /close-stmt/ (R908).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    close_stmt        Results of parsing the /close-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCloseStmt(tlist, close_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(CloseStmt), INTENT(OUT) :: close_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /close-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R908'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /close-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: no_spec        ! Flag to indicate UNIT= was absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /close-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! CLOSE...
    
    CALL ExpectKeyword(tlist, 1, scck_'CLOSE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! CLOSE(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /close-spec-list/
    
    list_count = ListItemCount(tlist(3:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      CALL parse_close_spec(  &
          tlist, &
          item_start, &
          item_finish,  &
          close_stmt,  &
          no_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error, but don't rely upon results of parse_close_spec.
      IF (.NOT. Failure(sub_err_list)) THEN
        IF ((item /= 1) .AND. no_spec) THEN
          ! Standard doesn't apply this constraint to a particular syntax rule.
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C908',  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='If the optional "UNIT=" is omitted, the /file-unit-number/ &
                &shall be the first item in the /close-spec-list/.' )
          ! Continue on error.
        END IF
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Check other constraints
    
    IF (.NOT. ALLOCATED(close_stmt%unit)) THEN
      ! Standard doesn't apply this constraint to a particular syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C908',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='A /file-unit-number/ shall be specified in a &
            &/close-spec-list/.' )
      ! Continue on error.
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseCloseStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /close-spec/ (R909).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /connect-spec/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /connect-spec/ in @a tlist.
  !!
  !! @param[in,out] close_stmt        Results of parsing the /close-spec/.  
  !! Note that this is an INTENT(INOUT) argument.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_close_spec( tlist, start, finish, close_stmt, no_spec,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(CloseStmt), INTENT(INOUT) :: close_stmt
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /close-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R909'
    
    ! Keywords for each /close-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(5) = [  &
        scck_'UNIT  ', scck_'IOSTAT', scck_'IOMSG ',  &
        scck_'ERR   ', scck_'STATUS' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /close-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No - this is a unit spec.
      no_spec = .TRUE.
      opt = 1
    ELSE
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/close-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(close_stmt%unit)) THEN
        CALL already('UNIT')
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber(  &
              tlist,  &
              start,  &
              finish,  &
              close_stmt%unit,  &
              sub_err_list )
        ELSE
          CALL ParseFileUnitNumber(  &
              tlist,  &
              assign + 1,  &
              finish,  &
              close_stmt%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! IOSTAT
      IF (ALLOCATED(close_stmt%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            close_stmt%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! IOMSG
      IF (ALLOCATED(close_stmt%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            close_stmt%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)      ! ERR
      IF (ALLOCATED(close_stmt%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(close_stmt%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            close_stmt%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (5)      ! STATUS
      IF (ALLOCATED(close_stmt%status)) THEN
        CALL already('STATUS')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            close_stmt%status,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /close-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      ! Standard doesn't specify a particular syntax rule for this constraint.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C907',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this /close-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_close_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of CloseStmt%Dump - write a debugging representation of 
  !! the results of parsing a close statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE close_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CloseStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%unit)) THEN
      WRITE (unit, "(A,'UNIT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%unit)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iomsg)) THEN
      WRITE (unit, "(A,'IOMSG = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iomsg)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iostat)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iostat)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%err)) THEN
      WRITE (unit, "(A,'ERR = ',I0)") pfx, .ValueOf. pnode%err
    END IF
    IF (ALLOCATED(pnode%status)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%status)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE close_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CloseStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE close_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CloseStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    ! err
    IF (ALLOCATED(pnode%status)) CALL pnode%status%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE close_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CloseStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE close_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CloseStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    ! err
    IF (ALLOCATED(pnode%status)) CALL pnode%status%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE close_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CloseStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'close-stmt'.
  
  FUNCTION close_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'close-stmt'
    
  END FUNCTION close_What
  
END MODULE CloseStmts
