! $Id: ComponentDefStmts.f90 2807 2019-03-22 23:17:22Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ComponentDefStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /data-component-def-stmt/'s and 
!! /proc-component-def-stmt/'s.
!!
!! A /data-component-def-stmt/ bears some resemblance to a simplified 
!! /type-declaration-stmt/, and we use some of the routines from the 
!! TypeDeclarationStmts module in this module.  In particular, 
!! we borrow the EntityDetails type to store the results of parsing 
!! each component.
!!
!! Unlike entity declarations, each component-def-stmt completely 
!! defines the component (data or procedure).  This means that we 
!! can check most of the constraints at Parse time.

MODULE ComponentDefStmts
  
  USE CompilerKinds
  USE DeclarationTypeSpecs
  USE Identifiers
  USE ParseNodes
  USE ProcedureDeclarationStmts
  USE SourceLocations
  USE SubscriptLists
  USE TypeDeclarationStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDataComponentDefStmt
  PUBLIC :: ParseProcComponentDefStmt
  
  ! Types required for calling public procedures.
  PUBLIC :: DeclTypeSpec      ! From DeclarationTypeSpecs.
  PUBLIC :: EntityDetails     ! From TypeDeclarationStmts.
  PUBLIC :: ProcDecl          ! From ProcedureDeclarationStmts.
  
  !-----------------------------------------------------------------------------
  
  !> Attributes in a /data-component-def-stmt/. 
  !!
  !! This is a simplified version of EntityAttrs.
  TYPE, EXTENDS(AttrBase), PUBLIC :: DataComponentAttrs
    LOGICAL :: allocatable = .FALSE.
    LOGICAL :: contiguous = .FALSE.
    LOGICAL :: pointer = .FALSE.
    !> See the module AccessStmts for more details.
    INTEGER :: access = 0
    
    !> Dimension specifications.
    !!
    !! If not allocated then the entity is a scalar.
    TYPE(DimSpec), ALLOCATABLE :: dim_specs(:)
    
    !> Codimension specification.
    !!
    !! If not allocated then the entity is not a coarray.
    TYPE(DimSpec), ALLOCATABLE :: codim_specs(:)
  CONTAINS
    PROCEDURE :: Dump => dca_Dump
    PROCEDURE :: Query => dca_Query
    PROCEDURE :: Modify => dca_Modify
    PROCEDURE, NOPASS :: What => dca_What
    PROCEDURE :: GetNatures => dca_GetNatures
  END TYPE DataComponentAttrs
  
  !> Attributes in a /proc-component-def-stmt/.
  TYPE, EXTENDS(AttrBase), PUBLIC :: ProcComponentAttrs
    !> Procedure interface.
    !!
    !! Not allocated if the statement had no interface specification.
    TYPE(ProcInterface), ALLOCATABLE :: proc_intf
    !> Procedure component has the pointer attribute.
    !!
    !! Always true after successful construction - C452 doesn't permit false.
    LOGICAL :: pointer = .FALSE.
    !> Indicates that the PASS attribute was provided.
    !!
    !! If /arg-name/ was provided within the PASS attribute specification, 
    !! then pass_arg_name will be allocated as well.
    LOGICAL :: pass = .FALSE.
    !> Allocated if PASS was specified and an /arg-name/ was specified within 
    !! the pass attribute specification.
    !!
    !! Otherwise (no PASS attribute, or no /arg-name/ specification) not 
    !! allocated.
    TYPE(Identifier), ALLOCATABLE :: pass_arg_name
    !> No pass attribute.
    LOGICAL :: nopass = .FALSE.
    !> Access specification.
    INTEGER :: access
  CONTAINS
    PROCEDURE :: Dump => pca_Dump
    PROCEDURE :: Query => pca_Query
    PROCEDURE :: Modify => pca_Modify
    PROCEDURE, NOPASS :: What => pca_What
    PROCEDURE :: Check => pca_Check
    PROCEDURE :: GetNatures => pca_GetNatures
  END TYPE ProcComponentAttrs
  
  !> Common parent type to use for component definitions.
  !!
  !! This is used in the type for the construct that holds the definitions.
  TYPE, EXTENDS(SCParseNode), ABSTRACT, PUBLIC :: ComponentDefStmt
  END TYPE ComponentDefStmt
  
  !> The results of parsing a /proc-component-def-stmt/.
  TYPE, EXTENDS(ComponentDefStmt), PUBLIC :: ProcComponentDefStmt
    !> The procedure component attributes.
    !!
    !! This is a pointer to simplify aliasing of the object with the 
    !! entity declaration objects.
    !!
    !! The ProcComponentDefStmt object owns the object that the pointer 
    !! is associated with.  Finalization is used to manage its lifetime.
    !!
    !! Always associated after succssful construction.
    TYPE(ProcComponentAttrs), POINTER :: attrs
    !> The procedure component declarations.
    TYPE(ProcDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => pcd_Dump
    PROCEDURE :: Query => pcd_Query
    PROCEDURE :: Modify => pcd_Modify
    PROCEDURE, NOPASS :: What => pcd_What
    
    FINAL :: pcd_Final_scalar
  END TYPE ProcComponentDefStmt
  
  !> The results of parsing a /data-component-def-stmt/.
  TYPE, EXTENDS(ComponentDefStmt), PUBLIC :: DataComponentDefStmt
    !> The /declaration-type-spec/ at the front of the statement.
    TYPE(DeclTypeSpec) :: decl_spec
    !> The component attributes.
    !!
    !! This is a pointer to simplify aliasing of the object with the 
    !! entity declaration objects.
    !!
    !! The DataComponentDefStmt object owns the object that the pointer 
    !! is associated with.  Finalization is used to manage its lifetime.
    !!
    !! Always associated after succssful construction.
    TYPE(DataComponentAttrs), POINTER :: attrs => NULL()
    !> The components declared by the statement.
    !!
    !! Always allocated, always size greater than zero.
    TYPE(EntityDetails), ALLOCATABLE :: names(:)
  CONTAINS
    PROCEDURE :: Dump => dcd_Dump
    PROCEDURE :: Query => dcd_Query
    PROCEDURE :: Modify => dcd_Modify
    PROCEDURE, NOPASS :: What  => dcd_What
    
    Final :: dcd_Final_scalar
  END TYPE DataComponentDefStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-component-def-stmt/ (R436).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    dcd_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDataComponentDefStmt(tlist, dcd_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DataComponentDefStmt), INTENT(OUT) :: dcd_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /data-component-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R436'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: finish         ! Syntax element final token.
    INTEGER :: it             ! Token index
    INTEGER :: declare        ! Location of the :: token.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! Name index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /data-component-def-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Locate the tokens associated with the declaration-type-spec.  This 
    ! always starts from token 1, but we need to work out where it finishes.
    
    finish = GetDeclarationTypeSpecFinish(tlist, .FALSE.)
    IF (finish == 0) finish = SIZE(tlist)
    
    CALL ParseDeclarationTypeSpec(  &
        tlist(:finish),  &
        .FALSE.,  &
        dcd_stmt%decl_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(err_list)) RETURN
    
    it = finish + 1
    
    !---------------------------------------------------------------------------
    ! Go looking for attributes.
    
    ALLOCATE(dcd_stmt%attrs)
    
    ! If we have attributes then there will be a comma after the type-spec.
    IF (Match(tlist, it, ittComma, scck_',')) THEN
      ! If we have attributes then we will have a bare :: somewhere in the 
      ! statement.
      declare = FindBare(tlist, ittDeclare, scck_'::', it)
      IF (declare == 0) THEN
        finish = SIZE(tlist)
      ELSE
        finish = declare - 1
      END IF
      CALL parse_dcd_attributes(  &
          tlist(it+1:finish),  &
          dcd_stmt%attrs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (declare == 0) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(finish)),  &
            MSG='Expected "::".' )
        RETURN
      END IF
      it = declare + 1
    ELSE
      ! The double colon is optional.
      IF (Match(tlist, it, ittDeclare, scck_'::')) it = it + 1
    END IF
    
    ! CLASS(xxx) requires POINTER or ALLOCATABLE.
    IF ( dcd_stmt%decl_spec%is_polymorphic  &
        .AND. .NOT. ( dcd_stmt%attrs%allocatable  &
          .OR. dcd_stmt%attrs%pointer ) ) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C407',  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='A component declared with CLASS must have the ALLOCATABLE or &
            &POINTER attributes.' )
      ! Carry on.
    END IF
    
    !---------------------------------------------------------------------------
    ! The /entity-decl-list/.
    
    ! it here should point to the token that is the start of the 
    ! variable name list.
    IF (dcd_stmt%attrs%allocatable .OR. dcd_stmt%attrs%POINTER) THEN
      CALL ParseDeclList(  &
          tlist(it:),  &
          dcd_stmt%attrs,  &
          .TRUE.,  &
          DeferredShapeSpecListMask,  &
          dcd_stmt%names,  &
          sub_err_list )
    ELSE
      CALL ParseDeclList(  &
          tlist(it:),  &
          dcd_stmt%attrs,  &
          .TRUE.,  &
          ExplicitShapeSpecListMask,  &
          dcd_stmt%names,  &
          sub_err_list )
    END IF
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Check constraints.
    
    ! Some constraints have already been checked (or enforced) through 
    ! the attribute parsing and subsequent limitations on what is acceptable 
    ! during component parsing.
    
    DO i = 1, SIZE(dcd_stmt%names)
      ! These tests might be redundant given the syntax constraints.  Commented 
      ! out for now.
      !IF (ALLOCATED(names(i)%dim_specs)) THEN
      !  ! If the component has the allocatable or pointer attribute then any
      !  ! /component-array-spec/ must be a /deferred-shape-spec-list/. If
      !  ! the component doesn't have those attributes then any 
      !  ! /component-array-spec/ must be an /explicit-shape-spec-list/.
      !  IF (attrs%allocatable .OR. attrs%pointer) THEN
      !    IF (.NOT. IsDeferredShapeSpecList(names(i)%dim_specs))  &
      !        CALL Add( err_list, CODE=errSyntax,  &
      !          COMPONENT='C441 via ' // comp,  &
      !          MSG='If the POINTER or ALLOCATABLE attribute is specified &
      !            &then each /component-array-spec/ shall be a &
      !            &/deferred-shape-spec-list/. ')
      !  ELSE
      !    IF (.NOT. IsExplicitShapeSpecList(names(i)%dim_specs))  &
      !        CALL Add( err_list, CODE=errSyntax,  &
      !          COMPONENT='C445 via ' // comp,  &
      !          MSG='If neither the POINTER or ALLOCATABLE attribute is &
      !            &specified then each /component-array-spec/ shall be a &
      !            &/explicit-shape-spec-list/. ')
      !  END IF
      !END IF
        
      IF (ALLOCATED(dcd_stmt%names(i)%codim_specs)) THEN
        ! If the component has a coarray-spec it shall be a 
        ! /deferred-coshape-spec-list/ and the component shall be 
        ! ALLOCATABLE.
        IF (.NOT. dcd_stmt%attrs%allocatable)  &
            CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C442 via ' // comp,  &
!              LOCATION=names(i)%location        
              MSG='If a /coarray-spec/ appears the component shall have &
                &the ALLOCATABLE attribute.' )
        IF (.NOT. IsDeferredCoshapeSpecList(dcd_stmt%names(i)%codim_specs))  &
            CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C442 via ' // comp,  &
!              LOCATION=names(i)%location        
              MSG='If a /coarray-spec/ appears it shall be a &
                &/deferred-coshape-spec-list/.' )
      END IF
      
      IF (dcd_stmt%attrs%contiguous) THEN
        IF (.NOT. ALLOCATED(dcd_stmt%names(i)%dim_specs))  &
            CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C448 via ' // comp,  &
!              LOCATION=names(i)%location        
              MSG='If the CONTIGUOUS attribute is specified the component &
                &shall be an array.' )
      END IF
    END DO
    
    ! We assume here that the CHARACTER intrinsic type cannot be renamed.
    IF (ALLOCATED(dcd_stmt%decl_spec%type_spec)) THEN
      IF (.NameOf.dcd_stmt%decl_spec%type_spec%type_name /= 'CHARACTER') THEN
        DO i = 1, SIZE(dcd_stmt%names)
          IF (ALLOCATED(dcd_stmt%names(i)%len)) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C449 on R438',  &
                LOCATION=dcd_stmt%names(i)%len%GetLocation(),  &
                MSG='A /char-length/ specifier is only permitted if the &
                  &entity being declared is of type CHARACTER.' )
          END IF
        END DO
      END IF
    END IF
    
  END SUBROUTINE ParseDataComponentDefStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentDefStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dcd_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataComponentDefStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables.
    
    INTEGER :: i              ! Name index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/declaration-type-spec/ = ')", ADVANCE='NO') pfx
    CALL WriteDeclTypeSpec(unit, pnode%decl_spec)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'attributes:')") pfx
    CALL pnode%attrs%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/component-decl-list/')") pfx
    DO i = 1, SIZE(pnode%names)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%names(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE dcd_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentDefStmt%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE dcd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DataComponentDefStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%decl_spec%Query(visitor)
    
    CALL pnode%attrs%Query(visitor)
    
    DO i = 1, SIZE(pnode%names)
      CALL pnode%names(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dcd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentDefStmt%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE dcd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DataComponentDefStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%decl_spec%Modify(visitor)
    
    CALL pnode%attrs%Modify(visitor)
    
    DO i = 1, SIZE(pnode%names)
      CALL pnode%names(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dcd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentDefStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name for the node.
  
  FUNCTION dcd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'data-component-def-stmt'
    
  END FUNCTION dcd_What
  
  
  !*****************************************************************************
  !!
  !> Scalar finalizer for a DataComponentDefStmt object.
  !!
  !! @param[in,out] dcd_stmt          The object to be finalized.
  
  SUBROUTINE dcd_Final_scalar(dcd_stmt)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DataComponentDefStmt), INTENT(INOUT) :: dcd_stmt
    
    !***************************************************************************
    
    IF (ASSOCIATED(dcd_stmt%attrs)) DEALLOCATE(dcd_stmt%attrs)
    
  END SUBROUTINE dcd_Final_scalar
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProcComponentDefStmt.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /proc-component-def-stmt/ (R440).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    pcd_stm           The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseProcComponentDefStmt(tlist, pcd_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProcComponentDefStmt), INTENT(OUT) :: pcd_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /proc-component-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R440'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token.
    INTEGER :: it_end         ! End of the current clause.    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: declare        ! Index of the declare token.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /proc-component-def-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! PROCEDURE(xxx) ...
    
    CALL ExpectKeyword(tlist, 1, scck_'PROCEDURE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, it_end, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! optional interface
    
    ALLOCATE(pcd_stmt%attrs)
    
    IF (it_end /= 3) THEN
      ALLOCATE(pcd_stmt%attrs%proc_intf)
      CALL ParseProcInterface(  &
          tlist(3:it_end-1),  &
          pcd_stmt%attrs%proc_intf,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      it = it_end + 1
    ELSE
      it = 4
    END IF
    
    !---------------------------------------------------------------------------
    ! Go looking for attributes.  POINTER is required, but we'll pretend it
    ! is not for now.
    
    CALL Expect(tlist, it, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    declare = FindBare(tlist, ittDeclare, scck_'::', it)
    IF (declare == 0) THEN
      it_end = SIZE(tlist)
    ELSE
      it_end = declare - 1
    END IF
    CALL parse_pcd_attributes(tlist(it:it_end), pcd_stmt%attrs, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    it = it_end + 1
    
    !---------------------------------------------------------------------------
    ! ::
    
    ! With correct synta we would have found the declare above, but to handle
    ! a missing declare repeat the check.
    CALL Expect(tlist, it, ittDeclare, scck_'::', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! /proc-decl-list/
    
    CALL ParseProcDeclList(  &
        tlist(it:),  &
        pcd_stmt%attrs,  &
        pcd_stmt%decls,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseProcComponentDefStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentDefStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE pcd_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcComponentDefStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables.
    
    INTEGER :: i              ! Declaration index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'attributes:')") pfx
    CALL pnode%attrs%Dump(unit, pfx // '  ')
    
    ! The declarators.
    WRITE (unit, "(A,'/proc-decl-list/')") pfx
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL WriteProcDecl(unit, pnode%decls(i))
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE pcd_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentDefStmt%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE pcd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcComponentDefStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%attrs%Query(visitor)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pcd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentDefStmt%Query - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE pcd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcComponentDefStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%attrs%Modify(visitor)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pcd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentDefStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name for the node.
  
  FUNCTION pcd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'proc-component-def-stmt'
    
  END FUNCTION pcd_What
  
  
  !*****************************************************************************
  !!
  !> Scalar finalizer for a ProcComponentDefStmt object.
  !!
  !! @param[in,out] pcd_stmt          The object to be finalized.
  
  SUBROUTINE pcd_Final_scalar(pcd_stmt)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProcComponentDefStmt), INTENT(INOUT) :: pcd_stmt
    
    !***************************************************************************
    
    IF (ASSOCIATED(pcd_stmt%attrs)) DEALLOCATE(pcd_stmt%attrs)
    
  END SUBROUTINE pcd_Final_scalar
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataComponentAttrs.
  
  
  !*****************************************************************************
  !!
  !> Parse the attributes part of a /data-component-def-stmt/.
  !!
  !! The attributes part is a list of /component-attr-spec/'s.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    attrs             Object describing the /component-attr/'s.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! @todo Consider holding over the parse of the DIMENSION attribute (if
  !! it is present) until we have determined the presence or otherwise
  !! of the ALLOCATABLE and POINTER attributes.
  
  SUBROUTINE parse_dcd_attributes(tlist, attrs, err_list)
    
    USE AccessStmts
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DataComponentAttrs), INTENT(OUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a single /component-attr-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R437'
    
    ! The list of /component-attr-spec/ keywords.
    CHARACTER(*,KIND=scck), PARAMETER :: attr_names(7) = [  &
        scck_'PUBLIC     ', scck_'PRIVATE    ', scck_'ALLOCATABLE',  &
        scck_'CODIMENSION', scck_'CONTIGUOUS ', scck_'DIMENSION  ',  &
        scck_'POINTER    ' ] 
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option.    
    INTEGER :: comma          ! Index of the next comma.
    INTEGER :: it_end         ! Index of the end of the current clause.
    
    ! Location of dimension attribute for error reporting.
    TYPE(SourceLocation) :: dim_loc
    
    ! Location of codimension attribute for error reporting.
    TYPE(SourceLocation) :: codim_loc
  
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    it = 1
    DO
      comma = FindBare(tlist, ittComma, scck_',', it) 
      IF (comma == 0) THEN
        it_end = SIZE(tlist)
      ELSE
        it_end = comma - 1
      END IF
      
      CALL ExpectKeyword( tlist, it, attr_names, '/component-attr-spec/',  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Process the attribute into the flags of the attrs object.
      CALL process_attribute
      
      ! No comma means we are done.
      IF (comma == 0) EXIT
      
      it = comma + 1
    END DO
    
    !---------------------------------------------------------------------------
    ! Constraint checking
    
    ! If we have a DIMENSION specification and a POINTER or ALLOCATABLE
    ! specification then the /component-array-spec/ must be a 
    ! /deferred-shape-spec-list/ by C441.  On the other hand, if we don't
    ! have POINTER or ALLOCATABLE, then the /component-array-spec/ must
    ! be an /explicit-shape-spec-list/ by C445.  This also enforces
    ! R439 which limits the options for a /component-array-spec/ - if we
    ! move the constraint checking then alternative checking will need to
    ! be put in. 
    
    IF (attrs%pointer .OR. attrs%allocatable) THEN
      IF (ALLOCATED(attrs%dim_specs)) THEN
        IF (.NOT. IsDeferredShapeSpecList(attrs%dim_specs))  &
            CALL Add( err_list, CODE=errSyntax,  &
                COMPONENT='C441 via ' // comp,  &
    !            LOCATION=dim_loc, ??
                MSG='If a component has either the ALLOCATABLE or POINTER &
                  &attribute then any /component-array-spec/ must be a &
                  &/deferred-shape-spec-list/.' )
      END IF
    ELSE
      IF (ALLOCATED(attrs%dim_specs)) THEN
        IF (.NOT. IsExplicitShapeSpecList(attrs%dim_specs))  &
            CALL Add( err_list, CODE=errSyntax,  &
                COMPONENT='C445 via ' // comp,  &
    !            LOCATION=dim_loc, ??
                MSG='If a component has either the ALLOCATABLE or POINTER &
                  &attribute then any /component-array-spec/ must be an &
                  &/explicit-shape-spec-list/.' )
      END IF
    END IF
    
    ! To be a coarray the component must be allocatable and the /coarray-spec/
    ! must be a /deferred-coshape-spec-list/.
    IF (ALLOCATED(attrs%codim_specs)) THEN
      IF (.NOT. attrs%allocatable)  &
          CALL Add( err_list, CODE=errSyntax,  &
              COMPONENT='C442 via ' // comp,  &
  !            LOCATION=codim_loc, ??
              MSG='If a /coarray-spec/ appears the component shall have the &
                &ALLOCATABLE attribute.' )
          
      IF (.NOT. IsDeferredCoshapeSpecList(attrs%codim_specs))  &
          CALL Add( err_list, CODE=errSyntax,  &
              COMPONENT='C442 via ' // comp,  &
  !            LOCATION=codim_loc, ??
              MSG='If a /coarray-spec/ appears it shall be a &
                &/deferred-coshape-spec-list/.' )
    END IF
    
  CONTAINS
    
    ! Just to make the structure of the main loop clearer.
    SUBROUTINE process_attribute
      INTEGER :: closing    ! Location of closing parenthesis.
      !*************************************************************************
      SELECT CASE (opt)
      CASE (1)  ! PUBLIC
        CALL do_access(iaccPublic)
        
      CASE (2)  ! PRIVATE
        CALL do_access(iaccPrivate)
        
      CASE (3)  ! ALLOCATABLE
        IF (attrs%allocatable) THEN
          CALL already('"ALLOCATABLE"')
        ELSE 
          IF (attrs%pointer) CALL pointer_and_allocatable
          attrs%allocatable = .TRUE.
          CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
        END IF
        
      CASE (4)  ! CODIMENSION
        IF (ALLOCATED(attrs%codim_specs)) THEN
          CALL already('"CODIMENSION"')
        ELSE
          CALL Expect(tlist, it + 1, ittDelimiter, scck_'[', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ExpectClosing(tlist, it + 1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          ! The type of coarray spec is checked below once we know about
          ! the allocatable attribute.
          CALL ParseCoarraySpec(tlist(it+1:closing), attrs%codim_specs,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! This location points to the CODIMENSION keyword.
          IF (.NOT. Failure(sub_err_list))  &
              codim_loc = QueryLocation(tlist(it))
          
          CALL ExpectEnd(tlist, closing + 1, comp, err_list)
        END IF
        
      CASE (5)  ! CONTIGUOUS
        IF (attrs%contiguous) THEN
          CALL already('"CONTIGUOUS"')
        ELSE
          attrs%contiguous = .TRUE.
          CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
        END IF
        
      CASE (6)  ! DIMENSION
        IF (ALLOCATED(attrs%dim_specs)) THEN
          CALL already('"DIMENSION"')
        ELSE
          CALL Expect(tlist, it+1, ittDelimiter, scck_'(', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ExpectClosing(tlist, it+1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          ! The type of array spec is checked below once we know about the 
          ! presence of the ALLOCATABLE and POINTER attributes.
          CALL ParseArraySpec( tlist(it:it_end), .TRUE., attrs%dim_specs,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! This location points to the DIMENSION keyword.
          IF (.NOT. Failure(sub_err_list)) dim_loc = QueryLocation(tlist(it))
          
          CALL ExpectEnd(tlist, closing + 1, comp, err_list)
        END IF
        
      CASE (7)  ! POINTER 
        IF (attrs%pointer) THEN
          CALL already('"POINTER"')
        ELSE 
          IF (attrs%allocatable) CALL pointer_and_allocatable
          attrs%pointer = .TRUE.
          CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
        END IF
        
      END SELECT
    END SUBROUTINE process_attribute
    
    SUBROUTINE do_access(iacc)
      !-------------------------------------------------------------------------
      ! Arguments
      INTEGER, INTENT(IN) :: iacc
      !*************************************************************************
      IF (attrs%access /= 0) THEN
        CALL already('An /access-spec/')
      ELSE
        attrs%access = iacc
      END IF
      CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
    END SUBROUTINE do_access
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C439 via ' // comp,  &
          LOCATION=QueryLocation(tlist(it)),  &
          MSG=text // ' has already been specified for this &
            &/component-def-stmt/.' )
    END SUBROUTINE already
    
    SUBROUTINE pointer_and_allocatable
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C447 via ' // comp,  &
          LOCATION=QueryLocation(tlist(it)),  &
          MSG='A component shall not have both the POINTER and ALLOCATABLE &
            &attributes.' )
    END SUBROUTINE pointer_and_allocatable
    
  END SUBROUTINE parse_dcd_attributes
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentAttrs%Dump - write a debugging 
  !! representation of the results of parsing the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dca_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataComponentAttrs), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (pnode%allocatable) WRITE (unit, "(A,'ALLOCATABLE')") pfx // '  '
    IF (pnode%contiguous) WRITE (unit, "(A,'CONTIGUOUS')") pfx // '  '
    IF (pnode%pointer) WRITE (unit, "(A,'POINTER')") pfx // '  '
    IF (pnode%access /= 0) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL WriteAccess(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%dim_specs)) THEN
      WRITE (unit, "(A,'DIMENSION (')", ADVANCE='NO') pfx // '  '
      CALL WriteDimSpec(unit, pnode%dim_specs)
      WRITE (unit, "(')')", ADVANCE='NO')
    END IF
    IF (ALLOCATED(pnode%codim_specs)) THEN
      WRITE (unit, "(A,'CODIMENSION [')", ADVANCE='NO') pfx // '  '
      CALL WriteDimSpec(unit, pnode%codim_specs)
      WRITE (unit, "(']')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE dca_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentAttrs%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE dca_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DataComponentAttrs), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dca_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentAttrs%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE dca_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DataComponentAttrs), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dca_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataComponentAttrs%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION dca_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'component-attr-spec'
    
  END FUNCTION dca_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%GetNatures - get the natures specified by 
  !! the attributes in the statement.
  !!
  !! @param[in]     attrs             The attributes of the statement.
  !!
  !! @param[out]    natures           The natures specified by the attributes.
  !!
  !! In addition to the attributes proper, a data component is always 
  !! considered an entity with type.  That nature is specified here.
  
  SUBROUTINE dca_GetNatures(attrs, natures)
    
    USE AccessStmts
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataComponentAttrs), INTENT(IN) :: attrs
    INTEGER, INTENT(OUT), ALLOCATABLE :: natures(:)
    
    !***************************************************************************
    
    natures = [NatureDefHasType, NatureDefDataComponent]
    
    IF (attrs%allocatable)  natures = [natures, NatureDefAllocatable]
    IF (attrs%contiguous)   natures = [natures, NatureDefContiguous]
    IF (attrs%pointer)      natures = [natures, NatureDefPointer]
    
    SELECT CASE (attrs%access)
    CASE (iaccPublic)   ; natures = [natures, NatureDefPublic]
    CASE (iaccPrivate)  ; natures = [natures, NatureDefPrivate]
    END SELECT
    
    IF (ALLOCATED(attrs%dim_specs)) THEN
      natures = [natures, NatureDefDimension]
    END IF
    
    IF (ALLOCATED(attrs%codim_specs)) THEN
      natures = [natures, NatureDefCodimension]
    END IF
    
  END SUBROUTINE dca_GetNatures
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProcComponentAttrs.
  
  
  !*****************************************************************************
  !!
  !> Parse a /proc-component-attr-spec-list/ - the attributes of a 
  !! /proc-component-def-stmt/.
  !! 
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in,out] attrs             The procedure attributes.  This 
  !! is INOUT because some of the attributes (interface name) may have been 
  !! already set by the calling code.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_pcd_attributes(tlist, attrs, err_list)
    
    USE AccessStmts
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProcComponentAttrs), INTENT(INOUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a single /proc-component-attr-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R441'
    
    ! The list of /proc-component-attr-spec/ keywords.
    CHARACTER(*,KIND=scck), PARAMETER :: attr_names(5) = [  &
        scck_'POINTER', scck_'PASS   ', scck_'NOPASS ',  &
        scck_'PUBLIC ', scck_'PRIVATE' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    INTEGER :: it_end         ! Index of the end of the current clause.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: comma          ! Index of the next comma.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    it = 1
    DO
      comma = FindBare(tlist, ittComma, scck_',', it)
      IF (comma == 0) THEN
        it_end = SIZE(tlist)
      ELSE
        it_end = comma - 1
      END IF
    
      CALL ExpectKeyword(  &
          tlist, it,  &
          attr_names,  &
          '/proc-component-attr-spec/',  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Process the attribute into the flags of the attrs object.
      CALL process_attribute
      
      ! No comma means we are done.
      IF (comma == 0) EXIT
      
      it = comma + 1
    END DO
    
  CONTAINS
    
    SUBROUTINE do_access(iacc)
      !-------------------------------------------------------------------------
      ! Arguments
      INTEGER, INTENT(IN) :: iacc
      !*************************************************************************      
      IF (attrs%access /= 0) THEN
        CALL already('An /access-spec/')
      ELSE
        attrs%access = iacc
      END IF
      CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
    END SUBROUTINE do_access
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C451 on ' // comp,  &
          LOCATION=QueryLocation(tlist(it)),  &
          MSG=text // ' has already been specified for this &
            &/proc-component-def-stmt/.' )
    END SUBROUTINE already
    
    SUBROUTINE process_attribute
      SELECT CASE (opt)
      CASE (1)  ! POINTER
        IF (attrs%pointer) THEN
          CALL already('"POINTER"')
        ELSE
          attrs%pointer = .TRUE.
          CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
        END IF
        
      CASE (2)  ! PASS
        IF (attrs%pass) THEN
          CALL already('"PASS"')
        ELSE IF (attrs%nopass) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C455 on ' // comp,  &
              LOCATION=QueryLocation(tlist(it)),  &
              MSG='"NOPASS" has already been specified for this &
                &/proc-component-def-stmt/.' )
        ELSE
          attrs%pass = .TRUE.
          
          IF (Match(tlist, it + 1, ittDelimiter, scck_'(')) THEN
            CALL ExpectClosing(tlist, it + 1, closing, err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (Failure(err)) RETURN
          
            CALL Expect(tlist, it + 2, ittName, err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (Failure(err)) RETURN
            
            attrs%pass_arg_name = IdentifierIfortBug(tlist(it+1))
            
            CALL ExpectEnd(tlist(:closing-1), it + 3, comp, err_list)
            CALL ExpectEnd(tlist(:it_end), closing + 1, comp, err_list)
          ELSE
            CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
          END IF
        END IF
        
      CASE (3)  ! NOPASS
        IF (attrs%nopass) THEN
          CALL already('"NOPASS"')
        ELSE IF (attrs%pass) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C455 via ' // comp,  &
              LOCATION=QueryLocation(tlist(it)),  &
              MSG='"PASS" has already been specified for this &
                &/proc-component-def-stmt/.' )
        ELSE
          attrs%nopass = .TRUE.
          CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
        END IF
        
      CASE (4)  ! PUBLIC
        CALL do_access(iaccPublic)
        
      CASE (5)  ! PRIVATE
        CALL do_access(iaccPrivate)
        
      END SELECT
    END SUBROUTINE process_attribute
    
  END SUBROUTINE parse_pcd_attributes
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentAttrs%Dump - write a debugging 
  !! representation of the results of parsing the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE pca_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcComponentAttrs), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (pnode%nopass) THEN
      WRITE (unit, "(A,'NOPASS')") pfx
    ELSE IF (pnode%pass) THEN
      IF (ALLOCATED(pnode%pass_arg_name)) THEN
        WRITE (unit, "(A,'PASS(',A,')')") pfx, .NameOf. pnode%pass_arg_name
      ELSE
        WRITE (unit, "(A,'PASS')") pfx
      END IF
    END IF
    IF (pnode%pointer) WRITE (unit, "(A,'POINTER')") pfx
    IF (pnode%access /= 0) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL WriteAccess(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    
    ! The procedure interface.
    IF (ALLOCATED(pnode%proc_intf)) THEN
      WRITE (unit, "(A,'proc interface ')", ADVANCE='NO') pfx
      IF (ALLOCATED(pnode%proc_intf%decl)) THEN
        WRITE (unit, "('decl = ')", ADVANCE='NO')
        CALL WriteDeclTypeSpec(unit, pnode%proc_intf%decl)
      END IF
      IF (ALLOCATED(pnode%proc_intf%name)) THEN
        WRITE (unit, "('name = ""',A,'""')", ADVANCE='NO')  &
            .NameOf. pnode%proc_intf%name
      END IF
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE pca_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentAttrs%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE pca_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcComponentAttrs), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%proc_intf)) CALL pnode%proc_intf%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pca_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentAttrs%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE pca_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcComponentAttrs), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%proc_intf)) CALL pnode%proc_intf%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pca_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentAttrs%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION pca_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'proc-component-attr-spec'
    
  END FUNCTION pca_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcComponentAttrs%Check.
  !!
  !! @param[in]     pnode             The parse node being checked.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE pca_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcComponentAttrs), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    LOGICAL :: is_implicit    ! Indicates implicit interface.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! C453 requires that procedures with implicit interfaces have NOPASS 
    ! specified.  A procedure with an implicit interface does not have 
    ! an /interface-name/ for its /proc-interface/.
    
    is_implicit = .NOT. ALLOCATED(pnode%proc_intf)
    IF (.NOT. is_implicit) THEN
      is_implicit = .NOT. ALLOCATED(pnode%proc_intf%name)
    END IF
    
    IF (is_implicit .AND. (.NOT. pnode%nopass)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C453 on R440',  &
          LOCATION=.LocationOf. pnode,  &
          MSG='If a procedure pointer component has an implicit interface &
            &(the /proc-interface/ is missing or is a &
            &/declaration-type-spec/ ) then NOPASS shall be specified.' )
    END IF
    
  END SUBROUTINE pca_Check
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%GetNatures - get the natures specified by 
  !! the attributes in the statement.
  !!
  !! @param[in]     attrs             The attributes of the statement.
  !!
  !! @param[out]    natures           The natures specified by the attributes.
  
  SUBROUTINE pca_GetNatures(attrs, natures)
    
    USE AccessStmts
    USE IntentSpecs
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcComponentAttrs), INTENT(IN) :: attrs
    INTEGER, INTENT(OUT), ALLOCATABLE :: natures(:)
    
    !***************************************************************************
    
    natures = [NatureDefProcComponent]
    
    IF (attrs%pointer)      natures = [natures, NatureDefPointer]
    
    SELECT CASE (attrs%access)
    CASE (iaccPublic)   ; natures = [natures, NatureDefPublic]
    CASE (iaccPrivate)  ; natures = [natures, NatureDefPrivate]
    END SELECT
    
  END SUBROUTINE pca_GetNatures
  
END MODULE ComponentDefStmts
