! $Id: CriticalConstructs.f90 2193 2016-06-06 13:01:31Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CriticalConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /critical-construct/'s.

MODULE CriticalConstructs
  
  USE ExecutionParts
  USE Identifiers
  USE Labels
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCriticalConstruct
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing critical construct.
  TYPE, EXTENDS(BlockStorage), PUBLIC :: CriticalConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The label on the end critical statement, if present, not allocated 
    !! otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => critical_Dump
    PROCEDURE :: Query => critical_Query
    PROCEDURE :: Modify => critical_Modify
    PROCEDURE, NOPASS :: What => critical_What
    PROCEDURE :: Advertise => critical_Advertise
  END TYPE CriticalConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for an /associate-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    crit_construct    The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCriticalConstruct(st, crit_construct, err_list)
    
    USE CriticalConstructStNodes
    USE CriticalStmts
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(CriticalConstructStNode), INTENT(INOUT) :: st
    TYPE(CriticalConstruct), INTENT(OUT) :: crit_construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening critical statement.
    
    CALL ParseCriticalStmt(  &
        st%first%tlist,  &
        crit_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Executable children.
    
    CALL ParseBlock(st%exec, crit_construct%children, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! The closing end critical statement.
    
    CALL ParseEndCriticalStmt(  &
        st%last%tlist,  &
        crit_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      crit_construct%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseCriticalConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of CriticalConstruct%Dump - write a debugging 
  !! representation of the results of parsing an critical construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE critical_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CriticalConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    WRITE (unit, "(A,'child executable statements:')") pfx
    CALL DumpBlock(pnode%children, unit, pfx // '  ')
    
  END SUBROUTINE critical_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CriticalConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE critical_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(CriticalConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL QueryBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE critical_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CriticalConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE critical_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(CriticalConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL ModifyBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE critical_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CriticalConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'critical-construct'.
  
  FUNCTION critical_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'critical-construct'
    
  END FUNCTION critical_What
  
  
  
  !*****************************************************************************
  !!
  !> Implementation of CriticalConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE critical_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CriticalConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE critical_Advertise
  
END MODULE CriticalConstructs