! $Id: CriticalStmts.f90 2146 2016-05-29 20:40:14Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CriticalStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /critical-stmt/'s and 
!! /end-critical-stmt/'s.

MODULE CriticalStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCriticalStmt
  PUBLIC :: ParseEndCriticalStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /critical-stmt/ (R811).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /critical-stmt/.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /critical-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCriticalStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /critical-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R811'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /critical-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave construct_name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! CRITICAL ...
    
    CALL ExpectKeyword(tlist, it, scck_'CRITICAL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Thats it!
    
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseCriticalStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-critical-stmt/ (R812).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-critical-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /critical-stmt/ if one was present, not allocated if a construct 
  !! name was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndCriticalStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-critical-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R812'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-critical-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDCRITICAL or END CRITICAL ...
    
    CALL ExpectSplittableKeyword(  &
        tlist, 1,  &
        scck_'END', scck_'CRITICAL',  &
        it,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist, it,  &
        construct_name,  &
        '/critical-stmt/',  &
        '/end-critical-stmt/',  &
        'C809 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndCriticalStmt
  
END MODULE CriticalStmts
