! $Id: DoStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DoStatements module.


!*******************************************************************************
!!
!> Procedures for parsing /label-do-stmt/'s, /nonlabel-do-stmt/'s, 
!! /cycle-stmt/'s and /end-do-stmt/'s (eventually).

MODULE DoStmts
  
  USE DoStmtsUtils
  USE Expressions
  USE Identifiers
  USE ForallStmts             ! For the ForallHeader type and procedures.
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: GetLabel          ! From DoStmtUtils.
  
  PUBLIC :: ParseCycleStmt
  PUBLIC :: ParseEndDoStmt
  PUBLIC :: ParseLabelDoStmt
  PUBLIC :: ParseNonlabelDoStmt
  
  !-----------------------------------------------------------------------------
  
  !> Abstract parent type for the results of parsing a /loop-control/.
  TYPE, EXTENDS(TermParseNode), ABSTRACT, PUBLIC :: LoopControl
  CONTAINS
    PROCEDURE(loop_Dump), DEFERRED :: Dump
    PROCEDURE, NOPASS :: What => loop_What
  END TYPE LoopControl
  
  ! Interfaces for deferred bindings in LoopControl.
  ABSTRACT INTERFACE
    !> Interface of LoopControl%Dump.
    !!
    !! @param[in]     loop              The loop control.
    !!
    !! @param[in]     unit              The logical unit.
    !!
    !! @param[in]     pfx               The prefix.
    SUBROUTINE loop_Dump(pnode, unit, pfx)
      IMPORT :: LoopControl
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(LoopControl), INTENT(IN) :: pnode
      INTEGER, INTENT(IN) :: unit
      CHARACTER(*), INTENT(IN) :: pfx
    END SUBROUTINE loop_Dump
  END INTERFACE
  
  !> Results of parsing the index form of control.
  TYPE, PUBLIC, EXTENDS(LoopControl) :: IndexLoopControl
    !> Name of the index variable.
    !!
    !! client code needs to confirm that this is a scalar integer expression.
    TYPE(Identifier) :: name
    !> The initial expression.
    !!
    !! Always allocated.
    !!
    !! Client code needs to confirm that this is a scalar integer expression.
    CLASS(Expression), ALLOCATABLE :: initial
    !> The terminal expression.
    !!
    !! Always allocated.
    !!
    !! Client code needs to confirm that this is a scalar integer expression.
    CLASS(Expression), ALLOCATABLE :: terminal
    !> The incrementation (step) expression.
    !!
    !! Not allocated if no step was provided.
    !!
    !! If allocated, then client code needs to confirm that this is a 
    !! scalar integer expression.
    CLASS(Expression), ALLOCATABLE :: incrementation
  CONTAINS
    PROCEDURE :: Dump => index_Dump
    PROCEDURE :: Write => index_Write
    PROCEDURE :: Query => index_Query
    PROCEDURE :: Modify => index_Modify
    PROCEDURE :: Advertise => index_Advertise
    PROCEDURE :: GetLocation => index_GetLocation
  END TYPE IndexLoopControl
  
  !> Results of parsing the WHILE(...) form of loop control.
  TYPE, PUBLIC, EXTENDS(LoopControl) :: WhileLoopControl
    !> The expression.
    !!
    !! Client code needs to confirm that this is a scalar logical expression.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => while_Dump
    PROCEDURE :: Write => while_Write
    PROCEDURE :: Query => while_Query
    PROCEDURE :: Modify => while_Modify
    PROCEDURE :: Advertise => while_Advertise
    PROCEDURE :: GetLocation => while_GetLocation
  END TYPE WhileLoopControl
  
  !> Results of parsing the CONCURRENT(...) form of loop control.
  TYPE, PUBLIC, EXTENDS(LoopControl) :: ConcurrentLoopControl
    !> The results of parsing the /forall-header/ that follows CONCURRENT.
    TYPE(ForallHeader) :: header
  CONTAINS
    PROCEDURE :: Dump => concurrent_Dump
    PROCEDURE :: Write => concurrent_Write
    PROCEDURE :: Query => concurrent_Query
    PROCEDURE :: Modify => concurrent_Modify
    PROCEDURE :: Advertise => concurrent_Advertise
    PROCEDURE :: GetLocation => concurrent_GetLocation
  END TYPE ConcurrentLoopControl
  
  !> The results of parsing a cycle statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CycleStmt
    !> The construct name, if present.  Not allocated if not present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
  CONTAINS
    PROCEDURE :: Dump => cycle_Dump
    PROCEDURE :: Query => cycle_Query
    PROCEDURE :: Modify => cycle_Modify
    PROCEDURE, NOPASS :: What => cycle_What
    PROCEDURE :: Advertise => cycle_Advertise
  END TYPE CycleStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /cycle-stmt/ (R831).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    cycle_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCycleStmt(tlist, cycle_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(CycleStmt), INTENT(OUT) :: cycle_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /cycle-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R831'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option return.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /cycle-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! CYCLE ...
    
    CALL ExpectKeyword(tlist, 1, scck_'CYCLE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /construct-name/ is optional.
    
    CALL Expect( tlist, 2, [ittEndStatement, ittName], [scck_'', scck_''],  &
        'end-of-statement or /construct-name/', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    IF (opt == 1) RETURN
    
    cycle_stmt%construct_name = IdentifierIfortBug(tlist(2))
    
    ! That's it!
    CALL ExpectEOS(tlist, 3, comp, err_list)
    
  END SUBROUTINE ParseCycleStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /label-do-stmt/ (R817).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /label-do-stmt/.
  !!
  !! @param[out]    loop_control      The form of loop control in the 
  !! statement, if present, or not allocated if the statement had no loop 
  !! control part.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /label-do-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    label             The label.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseLabelDoStmt( tlist, loop_control, construct_name,  &
      the_label, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE Labels
    USE MatchUtils
    USE ParseUtils
    USE Scanner               ! For ScanLabel
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Label), INTENT(OUT) :: the_label
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /label-do-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R816'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /label-do-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave construct_name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! DO ...
    
    CALL ExpectKeyword(tlist, it, scck_'DO', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The label.
    
    IF (it > SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /label/.' )
      RETURN
    END IF
    CALL ParseLabel(  &
        tlist,  &
        it,  &
        it,  &
        the_label,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That can be it - the loop control is optional.  If we bail now it is 
    ! just left unallocated to indicate that it wasn't there.
    IF (SIZE(tlist) == it) RETURN
    
    !---------------------------------------------------------------------------
    ! The loop control.
    
    ! This also handles the end of statement check.
    CALL parse_loop_control(  &
        tlist,  &
        it + 1, SIZE(tlist),  &
        loop_control,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseLabelDoStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /nonlabel-do-stmt/ (R817).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /nonlabel-do-stmt/.
  !!
  !! @param[out]    loop_control      The form of loop control in the 
  !! statement, if present, or not allocated if the statement had no loop 
  !! control part.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /nonlabel-do-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseNonlabelDoStmt(tlist, loop_control, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /nonlabel-do-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R817'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /nonlabel-do-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave construct_name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! DO ...
    
    CALL ExpectKeyword(tlist, it, scck_'DO', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That can be it - the loop control is optional.  If we bail now it is 
    ! just left unallocated to indicate that it wasn't there.
    IF (SIZE(tlist) == it) RETURN
    
    !---------------------------------------------------------------------------
    ! The loop control.
    
    ! This also handles the end of statement check.
    CALL parse_loop_control(  &
        tlist,  &
        it + 1, SIZE(tlist),  &
        loop_control,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseNonlabelDoStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-do-stmt/ (R821).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-do-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /do-stmt/ if one was present, not allocated if a construct 
  !! name was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndDoStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-do-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R821'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-do-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDDO or END DO ...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'DO', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/do-stmt/',  &
        '/end-do-stmt/',  &
        'C813 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndDoStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /loop-control/ (R818).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the /loop-control/ in 
  !! @a tlist.
  !!
  !! @param[in]     finish            End index of the /loop-control/ in 
  !! @a tlist.
  !!
  !! @param[out]    loop_control      The results of parsing the loop control.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_loop_control(tlist, start, finish, loop_control, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for /loop-control/.
    CHARACTER(*), PARAMETER :: comp = 'R818'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Procedures for matching the various types of loop control.
    TYPE(MatchProcList) :: procs(4)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (start > finish) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /loop-control/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Optional leading comma.
    
    ! Put the match_comma routine at the end of the array so that it is easy 
    ! to chop off if we need to match the index types again after the optional 
    ! comma.
    procs(1)%item => match_index
    procs(2)%item => match_while
    procs(3)%item => match_concurrent
    procs(4)%item => match_comma
    CALL Expect(  &
        tlist,  &
        start,  &
        procs,  &
        '/do-variable/ "=", "WHILE (", "CONCURRENT (" or ","',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Both branches set it to indicate where the loop control proper bit 
    ! starts.
    IF (opt == 4) THEN
      ! We came across the optional comma.  Have another dip without that 
      ! option.
      CALL Expect(  &
          tlist,  &
          start + 1,  &
          procs(:3),  &
          '/do-variable/ "=", "WHILE (" or "CONCURRENT ("',  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      it = start + 1
    ELSE
      it = start
    END IF
    
    ! opt here will have been set above to one of the loop control options, 
    ! it will have been set to the first token in the option (skipping the 
    ! comma if it was there).  The child parsing procedures can take it 
    ! as read that the tokens that they regard as significant were there.
    !
    ! To save typing we use an INOUT err_list argument.
    SELECT CASE (opt)
    CASE (1)                  ! Index style control.
      CALL parse_index_control(  &
          tlist,  &
          it, finish,  &
          comp,  &
          loop_control,  &
          err_list)
    CASE (2)                  ! While style control.
      CALL parse_while_control(  &
          tlist,  &
          it, finish,  &
          comp,  &
          loop_control,  &
          err_list)
    CASE (3)                  ! Concurrent style control.
      CALL parse_concurrent_control(  &
          tlist,  &
          it, finish,  &
          comp,  &
          loop_control,  &
          err_list)
    CASE DEFAULT
      STOP 'Bad Expect option returned in DoStmts::parse_loop_control.'
    END SELECT
    
  END SUBROUTINE parse_loop_control
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the index = form of /loop-control/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the /loop-control/ in 
  !! @a tlist (excludes optional leading ",").
  !!
  !! @param[in]     finish            End index of the /loop-control/ in 
  !! @a tlist.
  !!
  !! @param[in]     comp              Component to use for error reporting.
  !!
  !! @param[out]    loop_control      The results of parsing the loop 
  !! control.  Not allocated in the event of error.
  !!
  !! @param[in,out] err_list          List of errors.
  
  SUBROUTINE parse_index_control( tlist, start, finish, comp, loop_control,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! "Temporary" object that we build the result in before moving into the 
    ! loop_control argument.
    TYPE(IndexLoopControl), ALLOCATABLE :: tmp
    
    INTEGER :: comma          ! Index of the next comma in the token sequence.
    INTEGER :: it             ! Index of current token.
    INTEGER :: it_end         ! Index of last token of current part.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(tmp)
    
    ! First token is the variable, second token is the "=", both guaranteed 
    ! by match_index.  We need to save the index variable name though.
    tmp%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! The initial expression.
    
    ! Go looking for the end of the initial expression.
    comma = FindBare(tlist(:finish), ittComma, scck_',', start + 2)
    IF (comma == 0) THEN
      it_end = finish
    ELSE
      it_end = comma - 1
    END IF
    
    CALL ParseExpr(tlist, start + 2, it_end, tmp%initial, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    IF (comma == 0) THEN
      ! We repeat the "where is the comma" test, but this gives us a 
      ! consistent error message.
      CALL Expect(tlist(:finish), finish + 1, ittComma, scck_',', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      STOP 'Error handling logic failure in DoStmts::parse_index_control.'
    END IF
    
    it = comma + 1
    
    !---------------------------------------------------------------------------
    ! The terminal expression.
    
    ! Go looking for the end of the terminal expression.
    comma = FindBare(tlist(:finish), ittComma, scck_',', it)
    IF (comma == 0) THEN
      it_end = finish
    ELSE
      it_end = comma - 1
    END IF
    
    ! If there's no incrementation expression, then it_end == finish and the 
    ! expression parsing handles the check for extraneous following tokens.
    CALL ParseExpr(tlist, it, it_end, tmp%terminal, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! The incrementation expression is optional.
    IF (comma == 0) THEN
      CALL MOVE_ALLOC(tmp, loop_control)
      RETURN
    END IF
    
    it = comma + 1
    
    !---------------------------------------------------------------------------
    ! The incrementation (step) expression.
    
    ! The expression parsing also handles the check for extraneous following 
    ! tokens.
    CALL ParseExpr(tlist, it, finish, tmp%incrementation, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Swap our temporary over.
    
    CALL MOVE_ALLOC(tmp, loop_control)
    
  END SUBROUTINE parse_index_control
  
  
  !*****************************************************************************
  !!
  !> Implementation of LoopControl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'loop-control'.
  
  FUNCTION loop_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'loop-control'
    
  END FUNCTION loop_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in IndexLoopControl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl%Dump  - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     pnode             The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  !!
  !! @param[in]     pfx               The prefix.
  
  SUBROUTINE index_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IndexLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'name = ',A)") pfx, .NameOf. pnode%name
    WRITE (unit, "(A,'initial = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%initial)
    WRITE (unit, "()")
    WRITE (unit, "(A,'terminal = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%terminal)
    WRITE (unit, "()")
      
    IF (ALLOCATED(pnode%incrementation)) THEN
      WRITE(unit, "(A,'incrementation = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%incrementation)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE index_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl - write a debugging representation of 
  !! the loop control to a logical unit.
  !!
  !! @param[in]     pnode             The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  !!
  !! @param[in]     pfx               The prefix.
  
  SUBROUTINE index_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IndexLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('name = ',A)", ADVANCE='NO') .NameOf. pnode%name
    WRITE (unit, "(' initial = ')", ADVANCE='NO')
    CALL WriteExpr(unit, pnode%initial)
    WRITE (unit, "(' terminal = ')", ADVANCE='NO')
    CALL WriteExpr(unit, pnode%terminal)
      
    IF (ALLOCATED(pnode%incrementation)) THEN
      WRITE(unit, "(' incrementation = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%incrementation)
    END IF
    
  END SUBROUTINE index_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE index_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IndexLoopControl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%initial%Query(visitor)
    CALL pnode%terminal%Query(visitor)
    
    IF (ALLOCATED(pnode%incrementation)) THEN
      CALL pnode%incrementation%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE index_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the oepration.
  
  SUBROUTINE index_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IndexLoopControl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%initial%Modify(visitor)
    CALL pnode%terminal%Modify(visitor)
    
    IF (ALLOCATED(pnode%incrementation)) THEN
      CALL pnode%incrementation%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE index_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE index_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IndexLoopControl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureRefDataObject],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE index_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of IndexLoopControl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION index_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(IndexLoopControl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION index_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for WhileLoopControl.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the WHILE() form of /loop-control/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the /loop-control/ in 
  !! @a tlist (excludes optional leading ",").
  !!
  !! @param[in]     finish            End index of the /loop-control/ in 
  !! @a tlist.
  !!
  !! @param[in]     comp              Component to use for error reporting.
  !!
  !! @param[out]    loop_control      The results of parsing the loop control.
  !!
  !! @param[in,out] err_list          List of errors.
  
  SUBROUTINE parse_while_control( tlist, start, finish, comp, loop_control,  &
      err_list )
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! "Temporary" object that we build the result in before moving into the 
    ! loop_control argument.
    TYPE(WhileLoopControl), ALLOCATABLE :: tmp
    
    INTEGER :: closing        ! Index of the matching closing delimiter.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(tmp)
    
    ! First token is the WHILE keyword, second is the (.  Go looking for the 
    ! closing parentheses.
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The (logical) expression
    
    CALL ParseExpr(tlist, start + 2, closing - 1, tmp%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! There should be nothing after the closing parentheses.
    
    ! @todo ExpectEOS?
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
    !---------------------------------------------------------------------------
    ! Swap our temporary over.
    
    CALL MOVE_ALLOC(tmp, loop_control)
    
  END SUBROUTINE parse_while_control
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%Dump - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  !!
  !! @param[in]     pfx               The prefix.
  
  SUBROUTINE while_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhileLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'while expression = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, "()")
    
  END SUBROUTINE while_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%Write - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE while_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhileLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('while expression = ')", ADVANCE='NO')
    CALL WriteExpr(unit, pnode%expr)
    
  END SUBROUTINE while_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE while_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(WhileLoopControl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE while_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE while_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(WhileLoopControl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE while_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE while_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhileLoopControl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE while_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhileLoopControl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION while_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(WhileLoopControl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%expr
    
  END FUNCTION while_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ConcurrentLoopControl.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the CONCURRENT() form of 
  !! /loop-control/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the /loop-control/ in 
  !! @a tlist (excludes optional leading ",").
  !!
  !! @param[in]     finish            End index of the /loop-control/ in 
  !! @a tlist.
  !!
  !! @param[in]     comp              Component to use for error reporting.
  !!
  !! @param[out]    loop_control      The results of parsing the loop control.
  !!
  !! @param[in,out] err_list          List of errors.
  
  SUBROUTINE parse_concurrent_control( tlist, start, finish, comp,  &
      loop_control, err_list )
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CLASS(LoopControl), INTENT(OUT), ALLOCATABLE :: loop_control
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! "Temporary" object that we build the result in before moving into the 
    ! loop_control argument.
    TYPE(ConcurrentLoopControl), ALLOCATABLE :: tmp
    
    INTEGER :: closing        ! Index of the matching closing delimiter.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(tmp)
    
    ! First token is the CONCURRENT keyword, second is the (.  Go looking 
    ! for the closing parentheses.
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Parse the forall header.
    
    CALL ParseForallHeader( tlist, start + 2, closing - 1, tmp%header,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! There should be nothing after the closing parentheses.
    
    ! @todo ExpectEOS?
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
    !---------------------------------------------------------------------------
    ! Swap our temporary over.
    
    CALL MOVE_ALLOC(tmp, loop_control)
    
  END SUBROUTINE parse_concurrent_control
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrrentLoopControl - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     pnode             The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  !!
  !! @param[in]     pfx               The prefix.
  
  SUBROUTINE concurrent_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConcurrentLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'forall-header = ')", ADVANCE='NO') pfx
    CALL pnode%header%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE concurrent_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrrentLoopControl%Write - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     pnode             The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE concurrent_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConcurrentLoopControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('forall-header = ')", ADVANCE='NO')
    CALL pnode%header%Write(unit)
    
  END SUBROUTINE concurrent_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrentLoopControl%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE concurrent_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ConcurrentLoopControl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! @todo
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE concurrent_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrentLoopControl%Modifyy.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE concurrent_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ConcurrentLoopControl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! @todo
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE concurrent_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrentLoopControl%Advertise - advertise the 
  !! identifiers referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE concurrent_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConcurrentLoopControl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE concurrent_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConcurrentLoopControl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION concurrent_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ConcurrentLoopControl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%header
    
  END FUNCTION concurrent_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Loop control match procedures.
  !
  ! These all search for the minimum set of tokens to decide on a particular 
  ! form of loop control (helped by the syntax restrictions that the index 
  ! variable in the index form of control must be a simple name).
  
  
  !*****************************************************************************
  !!
  !> Match function for a comma token.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token to start the match.
  !!
  !! @returns True if the token at index @a start in @a tlist is a comma.
  
  FUNCTION match_comma(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittComma, scck_',')
    
  END FUNCTION match_comma
  
  
  !*****************************************************************************
  !!
  !> Match function for an do index variable style control.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token to start the match.
  !!
  !! @returns .TRUE. if the token sequence commencing at start consists of 
  !! "name = ", false otherwise.
  
  FUNCTION match_index(tlist, start) RESULT(b)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = MatchName(tlist, start) .AND. Match(tlist, start + 1, ittAssign)
    
  END FUNCTION match_index
  
  
  !*****************************************************************************
  !!
  !> Match function for the DO WHILE style control.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token to start the match.
  !!
  !! @returns .TRUE. if the token sequence commencing at start consists of 
  !! "WHILE (", false otherwise.
  
  FUNCTION match_while(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'WHILE')  &
        .AND. Match(tlist, start + 1, ittDelimiter, scck_'(')
    
  END FUNCTION match_while
  
  
  !*****************************************************************************
  !!
  !> Match function for the DO CONCURRENT style control.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token to start the match.
  !!
  !! @returns .TRUE. if the token sequence commencing at start consists of 
  !! "CONCURRENT (", false otherwise.
  
  FUNCTION match_concurrent(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'CONCURRENT')  &
        .AND. Match(tlist, start + 1, ittDelimiter, scck_'(')
    
  END FUNCTION match_concurrent
  
  
  !*****************************************************************************
  !!
  !> Implementation of CycleStmt%Dump - write a debugging representation 
  !! of the results of parsing a stop statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE cycle_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CycleStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'/construct-name/ = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
  END SUBROUTINE cycle_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CycleStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cycle_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CycleStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cycle_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CycleStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cycle_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CycleStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cycle_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CycleStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'cycle-stmt'.
  
  FUNCTION cycle_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'cycle-stmt'
    
  END FUNCTION cycle_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of CycleStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE cycle_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CycleStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureRefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE cycle_Advertise
  
END MODULE DoStmts
