! $Id: ExitStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ExitStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /exit-stmt/'s.

MODULE ExitStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseExitStmt
  
  !-----------------------------------------------------------------------------
  
  !> Represents the results of parsing an exit statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ExitStmt
    !> /construct-name/ if present, otherwise not allocated.
    TYPE(Identifier), ALLOCATABLE :: construct_name
  CONTAINS
    PROCEDURE :: Dump => exit_Dump
    PROCEDURE :: Query => exit_Query
    PROCEDURE :: Modify => exit_Modify
    PROCEDURE, NOPASS :: What => exit_What
    PROCEDURE :: Advertse => exit_Advertise
  END TYPE ExitStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /exit-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    exit_stmt         The result of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseExitStmt(tlist, exit_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ExitStmt), INTENT(OUT) :: exit_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /exit-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R850'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option return.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /exit-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! EXIT ...
    
    CALL ExpectKeyword(tlist, 1, scck_'EXIT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /construct-name/ is optional.
    
    CALL Expect( tlist, 2, [ittEndStatement, ittName], [scck_'', scck_''],  &
        'end-of-statement or /construct-name/', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    IF (opt == 1) RETURN
    
    exit_stmt%construct_name = IdentifierIfortBug(tlist(2))
    
    ! That's it!
    CALL ExpectEOS(tlist, 3, comp, err_list)
    
  END SUBROUTINE ParseExitStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExitStmt%Dump - write a debugging representation 
  !! of the results of parsing a stop statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE exit_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExitStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'/construct-name/ = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
  END SUBROUTINE exit_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExitStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE exit_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExitStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE exit_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExitStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE exit_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExitStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE exit_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExitStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'exit-stmt'.
  
  FUNCTION exit_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'exit-stmt'
    
  END FUNCTION exit_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExitStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE exit_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExitStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureRefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE exit_Advertise
  
END MODULE ExitStmts
