! $Id: Expressions.f90 2169 2016-06-02 10:00:57Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Expressions module.


!*******************************************************************************
!!
!> Defines types for storing expressions.
!!
!! Because expressions are used everywhere it is easy to end up with cyclic 
!! dependencies between modules.  So the expression parsing routine is 
!! actually an external routine.

MODULE Expressions
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseExpr
  PUBLIC :: WriteExpr
  
  !-----------------------------------------------------------------------------
  
  !> Represents an expression.
  !!
  !! This should have been called Expr if we following the convention of 
  !! naming the type after the syntax rule.  But expr is a very handy 
  !! variable name.
  !!
  !! The Expression extensions are basically just storage types.
  TYPE, PUBLIC, EXTENDS(TermParseNode), ABSTRACT :: Expression
  CONTAINS
    PROCEDURE :: Dump => expr_Dump
    PROCEDURE, NOPASS :: What => expr_What
  END TYPE Expression
  
  !> An expression that consists of a unary operator.
  TYPE, EXTENDS(Expression), PUBLIC :: UnaryOpExpr
    TYPE(Identifier) :: op
    CLASS(Expression), ALLOCATABLE :: rhs
  CONTAINS
    PROCEDURE :: Write => unary_Write
    PROCEDURE :: GetLocation => unary_GetLocation
    PROCEDURE :: Query => unary_Query
    PROCEDURE :: Modify => unary_Modify
    PROCEDURE :: Advertise => unary_Advertise
  END TYPE UnaryOpExpr
  
  !> An expression that consists of a binary operator.
  !!
  !! One of lhs or rhs will be allocated.  If only one is allocated, then 
  !! op is not defined.  If both are allocated, op is the operator.
  TYPE, EXTENDS(Expression), PUBLIC :: BinaryOpExpr
    CLASS(Expression), ALLOCATABLE :: lhs
    TYPE(Identifier) :: op
    CLASS(Expression), ALLOCATABLE :: rhs
  CONTAINS
    PROCEDURE :: Write => binary_Write
    PROCEDURE :: GetLocation => binary_GetLocation
    PROCEDURE :: Query => binary_Query
    PROCEDURE :: Modify => binary_Modify
    PROCEDURE :: Advertise => binary_Advertise
  END TYPE BinaryOpExpr
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /expression/.
  !!
  !! To avoid cyclic dependencies this is implemented by an external 
  !! subroutine.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /expression/.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /expression/.
  !!
  !! @param[out]    ex                The results of parsing the expression.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! Expressions can contain other expressions, so recursive.
  RECURSIVE SUBROUTINE ParseExpr(tlist, start, finish, ex, err_list)
    
    USE Errors
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTERFACE
      RECURSIVE SUBROUTINE ParseExprExternal(tlist, start, finish, ex, err_list)
        USE Errors
        USE Tokens
        IMPORT :: Expression
        IMPLICIT NONE
        !-------------------------------------------------------------------------
        TYPE(Token), INTENT(IN) :: tlist(:)
        INTEGER, INTENT(IN) :: start
        INTEGER, INTENT(IN) :: finish
        CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
        TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
      END SUBROUTINE ParseExprExternal
    END INTERFACE
    
    !***************************************************************************
    
    CALL ParseExprExternal(tlist, start, finish, ex, err_list)
    
  END SUBROUTINE ParseExpr
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of an expression to a logical unit.
  !!
  !! @param[in]     unit              Local unit connected for formatted 
  !! output.
  !!
  !! @param[in]     ex                The expression to write.
  
  RECURSIVE SUBROUTINE WriteExpr(unit, ex)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    CLASS(Expression), INTENT(IN) :: ex
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTERFACE
      RECURSIVE SUBROUTINE WriteExprExternal(unit, ex)
        IMPORT :: Expression
        IMPLICIT NONE
        !-------------------------------------------------------------------------
        INTEGER, INTENT(IN) :: unit
        CLASS(Expression), INTENT(IN) :: ex
      END SUBROUTINE WriteExprExternal
    END INTERFACE
    
    !***************************************************************************
    
    CALL WriteExprExternal(unit, ex)
    
  END SUBROUTINE WriteExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of Expression%Dump - write a debugging representation 
  !! of an expression to a logical unit.
  !!
  !! @param[in]     pnode             The expression to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  !!
  !! Typically expressions are written inline in a single record, rather 
  !! than being dumped to a range of records, using the Write binding.  But 
  !! to give some flexibility we also have the Dump binding - similar to 
  !! SCParseNodes.
  
  SUBROUTINE expr_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Expression), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL pnode%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE expr_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of Expression%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'expr'
  !!
  !! We may specialize this for the extensions of Expression at a later 
  !! date.
  
  FUNCTION expr_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'expr'
    
  END FUNCTION expr_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in UnaryOpExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnaryOpExpr%Write - write a debugging representation 
  !! of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE unary_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnaryOpExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A,' {')", ADVANCE='NO') .NameOf. pnode%op
    CALL pnode%rhs%Write(unit)
    WRITE (unit, "('}')", ADVANCE='NO')
    
  END SUBROUTINE unary_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnaryOpExpr%GetLocation - get the location 
  !! in source of the start of the expression.
  !!
  !! @param[in]     expr              The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We can easily retrieve the location from the stored token.
  
  FUNCTION unary_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(UnaryOpExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%op%GetLocation()
    
  END FUNCTION unary_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnaryOpExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Unary expressions can contain other unary expressions, hence RECURSIVE.
  RECURSIVE SUBROUTINE unary_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnaryOpExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%rhs%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE unary_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnaryOpExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Unary expressions can contain other unary expressions, hence RECURSIVE.
  RECURSIVE SUBROUTINE unary_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnaryOpExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%rhs%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE unary_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnaryOpExpr%Advertise - advertise the identifiers 
  !! referenced or defined by the expression.
  !!
  !! @param[in]     expr              The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE unary_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnaryOpExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%op,  &
        NATURES=[NatureRefOperator],  &
        LOCATIONS=[.LocationOf. pnode%op] )
    
  END SUBROUTINE unary_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in BinaryOpExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of BinaryOpExpr%Write - write a debugging representation 
  !! of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE binary_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BinaryOpExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('{')", ADVANCE='NO')
    CALL pnode%lhs%Write(unit)
    WRITE (unit, "('} ',A,' {')", ADVANCE='NO') .NameOf. pnode%op
    CALL pnode%rhs%Write(unit)
    WRITE (unit, "('}')", ADVANCE='NO')
    
  END SUBROUTINE binary_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of BinaryOpExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     expr              The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We just forward the call to the left hand side expression.
  
  FUNCTION binary_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(BinaryOpExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%lhs%GetLocation()
    
  END FUNCTION binary_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of BinaryOpExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Binary expressions can contain other binary expressions, hence RECURSIVE.
  RECURSIVE SUBROUTINE binary_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BinaryOpExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%lhs%Query(visitor)
    CALL pnode%rhs%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE binary_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BinaryOpExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Binary expressions can contain other binary expressions, hence RECURSIVE.
  RECURSIVE SUBROUTINE binary_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BinaryOpExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%lhs%Modify(visitor)
    CALL pnode%rhs%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE binary_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BinaryOpExpr%Advertise - advertise the identifiers 
  !! referenced or defined by the expression.
  !!
  !! @param[in]     expr              The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE binary_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BinaryOpExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%op,  &
        NATURES=[NatureRefOperator],  &
        LOCATIONS=[.LocationOf. pnode%op] )
    
  END SUBROUTINE binary_Advertise
  
END MODULE Expressions
