! $Id: ForallStmts.f90 2803 2019-03-22 19:36:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ForallStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /forall-stmt/'s (R759), 
!! /forall-construct-stmt/'s, and /end-forall-stmt/'s.
!!
!! Also a procedure to parse the /forall-header/, which is also used 
!! by DO CONCURRENT statement parsing.

MODULE ForallStmts
  
  USE DeclarationTypeSpecs
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseForallStmt
  PUBLIC :: ParseForallConstructStmt
  PUBLIC :: ParseForallHeader
  PUBLIC :: ParseEndForallStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /forall-triplet-spec/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ForallTripletSpec
    !> Name of the index variable.
    TYPE(Identifier) :: index
    !> Starting limit.
    CLASS(Expression), ALLOCATABLE :: start
    !> Final limit.
    CLASS(Expression), ALLOCATABLE :: finish
    !> Step.
    !!
    !! Not allocated if the step was not provided.
    CLASS(Expression), ALLOCATABLE :: step
  CONTAINS
    PROCEDURE :: Write => foralltriplet_Write
    PROCEDURE, NOPASS :: What => foralltriplet_What
    PROCEDURE :: Query => foralltriplet_Query
    PROCEDURE :: Modify => foralltriplet_Modify
    PROCEDURE :: GetLocation => foralltriplet_GetLocation
  END TYPE ForallTripletSpec
  
  !> Stores the results of parsing a /forall-header/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ForallHeader
    !> Optional leading type spec - if not allocated the type spec was not 
    !! present.
    TYPE(TypeSpec), ALLOCATABLE :: type_spec
    !> The list of triplet specifications.
    !!
    !! Always allocated, size greater than zero.
    TYPE(ForallTripletSpec), ALLOCATABLE :: triplets(:)
    !> The scalar mask expression.
    !!
    !! Not allocated if no mask expression was present.
    CLASS(Expression), ALLOCATABLE :: mask
  CONTAINS
    PROCEDURE :: Write => forallhdr_Write
    PROCEDURE, NOPASS :: What => forallhdr_What
    PROCEDURE :: Query => forallhdr_Query
    PROCEDURE :: Modify => forallhdr_Modify
    PROCEDURE :: GetLocation => forallhdr_GetLocation
  END TYPE ForallHeader
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /forall-stmt/ (R759).  This is the 
  !! single line form of forall statement.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /forall-construct-stmt/.
  !!
  !! @param[out]    header            The results of parsing the 
  !! /forall-header/.
  !!
  !! @param[out]    child_tlist       List of tokens that make up the child 
  !! /action-stmt/ of the /forall-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseForallStmt(tlist, header, child_tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ForallHeader), INTENT(OUT) :: header
    TYPE(Token), INTENT(OUT), ALLOCATABLE :: child_tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /forall-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R759'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /forall-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! FORALL ....
    
    CALL Expect(tlist, 1, ittName, scck_'FORALL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! FORALL(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /forall-header/.
    
    CALL ParseForallHeader(tlist, 3, closing - 1, header, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! The child assignment statement.
    
    IF (SIZE(tlist) == closing) THEN
      ! The error message is cognisant of classification options.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(closing)),  &
          MSG='Expected /forall-assignment-stmt/ after FORALL (...).' )
      RETURN
    END IF
    
    child_tlist = tlist(closing+1:)
    
  END SUBROUTINE ParseForallStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /forall-construct-stmt/ (R750).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /forall-construct-stmt/.
  !!
  !! @param[out]    header            The results of parsing the 
  !! /forall-header/.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /forall-construct-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseForallConstructStmt(tlist, header, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ForallHeader), INTENT(OUT) :: header
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /forall-construct-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R751'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /forall-construct-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! FORALL ....
    
    CALL ExpectKeyword(tlist, it, scck_'FORALL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! FORALL(...)
    
    CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /forall-header/.
    
    CALL ParseForallHeader(tlist, it + 2, closing - 1, header, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Thats it!
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseForallConstructStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-forall-stmt/ (R758).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-forall-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /forall-construct-stmt/ if one was present, not allocated if a construct 
  !! name was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndForallStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-forall-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R758'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-forall-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDFORALL or END FORALL ...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'FORALL', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/forall-construct-stmt/',  &
        '/end-forall-stmt/',  &
        'C734 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndForallStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /forall-header/ (R751).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! parent statement.
  !!
  !! @param[in]     start             Start index of the header in @a tlist.
  !!
  !! @param[in]     finish            End index of the header in @a tlist.
  !!
  !! @param[out]    header            The results of parsing the header.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseForallHeader(tlist, start, finish, header, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ForallHeader), INTENT(OUT) :: header
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /forall-header/.
    CHARACTER(*), PARAMETER :: comp = 'R752'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: colon          ! Index of the double colon after the type spec.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: it             ! Current token index.
    INTEGER :: item           ! Current item index.
    INTEGER :: list_count     ! Number of items in the triplet spec list.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    INTEGER :: equals         ! Index of the equals.
    LOGICAL :: found_mask     ! Flag to indicate that a mask has been found.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /forall-header/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Optional type spec.
    
    ! Indicated by presence of a double colon.
    colon = FindBare(tlist, ittDeclare, scck_'::', start)
    IF (colon /= 0) THEN
      ALLOCATE(header%type_spec)
      CALL ParseTypeSpec(tlist(start:colon-1), header%type_spec, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Carry on in the event of failure.
      it = colon + 1
    ELSE
      it = start
    END IF
    
    !---------------------------------------------------------------------------
    ! List of triplet specs, optionally terminated by a mask.
    
    found_mask = .FALSE.
    
    item_finish = 0
    list_count = 0            ! Number of triplet specs.
    item = 0
    DO
      CALL GetNextListItem(tlist, it, finish, item_start, item_finish)
      IF (item_start == 0) EXIT
      
      item = item + 1
      
      ! We can pick the mask by the absence of the '='.
      equals = FindBare(tlist(item_start:item_finish), ittAssign, scck_'=')
      
      ! The mask must be the last thing in the list.
      IF (found_mask) THEN
        IF (equals /= 0) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='The /forall-triplet-spec/''s must appear ahead of the &
                &/scalar-mask-expr/.' )
        ELSE
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='Unexpected items in the /forall-header/ after the &
                &/scalar-mask-expr/.' )
          ! Give up on further processing of the header.
        END IF
        RETURN
      END IF
      
      IF (equals == 0) THEN
        ! This must be the mask.
        found_mask = .TRUE.
        CALL ParseExpr(  &
            tlist,  &
            item_start, item_finish,  &
            header%mask,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ELSE
        ! This must be a triplet spec.
        list_count = list_count + 1
      END IF
    END DO
    
    IF (list_count == 0) THEN
      ! There must be at least one triplet spec.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          MSG='Expected at least one /forall-triplet-spec/ in the &
            &/forall-header.' )
      RETURN
    END IF
    
    ! Now we go back and process the triplets.
    ALLOCATE(header%triplets(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, finish, item_start, item_finish)
      CALL parse_triplet_spec(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          header%triplets(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END DO
    
  END SUBROUTINE ParseForallHeader
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /forall-triplet-spec/ (R753).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the triplet in @a tlist.
  !!
  !! @param[in]     finish            End index of the triplet in @a tlist.
  !!
  !! @param[out]    triplet           The results of parsing a 
  !! /forall-triplet-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_triplet_spec(tlist, start, finish, triplet, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ForallTripletSpec), INTENT(OUT) :: triplet
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /forall-triplet-spec/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R753'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: colon          ! Index of the colon.
    TYPE(Error) :: err        ! Error code for child calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: it_end         ! Index of last token of current part.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /forall-triplet-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! name = 
    
    CALL ExpectName(tlist, start, '/index-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    triplet%index = IdentifierIfortBug(tlist(start))
    
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! The lower limit.
    
    colon = FindBare(tlist, ittColon, scck_':', start + 3)
    IF (colon > finish) colon = 0
    IF (colon == 0) THEN
      it_end = finish
    ELSE
      it_end = colon - 1
    END IF
    CALL ParseExpr(tlist, start + 2, it_end, triplet%start, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! The upper limit
    
    IF (colon == 0) THEN
      CALL Add(err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          MSG='Expected /forall-limit/.' )
      RETURN
    END IF
    it = colon + 1
    
    ! Because the step is optional, the upper limit may either finish 
    ! at the end of the token sequence for this syntax component (finish) 
    ! or before the next colon.
    colon = FindBare(tlist, ittColon, scck_':', it)
    IF (colon > finish) colon = 0
    IF (colon == 0) THEN
      it_end = finish
    ELSE
      it_end = colon - 1
    END IF
    CALL ParseExpr(tlist, it, it_end, triplet%finish, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Step is optional.
    IF (colon == 0) RETURN
    it = colon + 1
    
    !---------------------------------------------------------------------------
    ! The step.
    
    CALL ParseExpr(tlist, it, finish, triplet%step, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_triplet_spec
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ForallHeader
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallHeader%Write - write a debugging representation 
  !! of the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE forallhdr_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallHeader), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Triplet index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL WriteTypeSpec(unit, pnode%type_spec)
      WRITE (unit, "(' :: ')", ADVANCE='NO')
    END IF
    
    DO i = 1, SIZE(pnode%triplets)
      CALL pnode%triplets(i)%Write(unit)
      IF (i /= SIZE(pnode%triplets))  &
          WRITE (unit, "(' , ')", ADVANCE='NO')
    END DO
    
    IF (ALLOCATED(pnode%mask)) THEN
      WRITE (unit, "(' , ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%mask)
    END IF
    
  END SUBROUTINE forallhdr_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallHeader%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'forall-header'
  !!
  !! We may specialize this for the extensions of Expression at a later 
  !! date.
  
  FUNCTION forallhdr_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'forall-header'
    
  END FUNCTION forallhdr_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallHeader%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE forallhdr_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallHeader), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Triplet index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Query(visitor)
    END IF
    
    DO i = 1, SIZE(pnode%triplets)
      CALL pnode%triplets(i)%Query(visitor)
    END DO
    
    IF (ALLOCATED(pnode%mask)) THEN
      CALL pnode%mask%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE forallhdr_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallHeader%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE forallhdr_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallHeader), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Triplet index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Modify(visitor)
    END IF
    
    DO i = 1, SIZE(pnode%triplets)
      CALL pnode%triplets(i)%Modify(visitor)
    END DO
    
    IF (ALLOCATED(pnode%mask)) THEN
      CALL pnode%mask%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE forallhdr_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallHeader%GetLocation - get the start of 
  !! the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION forallhdr_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ForallHeader), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      loc = .Locationof. pnode%type_spec
    ELSE
      loc = .LocationOf. pnode%triplets(1)
    END IF
    
  END FUNCTION forallhdr_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ForallTripletSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallTripletSpec%Dump - dump a debugging 
  !! representation of the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE foralltriplet_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallTripletSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A,' = ')", ADVANCE='NO') .NameOf. pnode%index
    CALL WriteExpr(unit, pnode%start)
    WRITE (unit, "(' : ')", ADVANCE='NO')
    CALL WriteExpr(unit, pnode%finish)
    IF (ALLOCATED(pnode%step)) THEN
      WRITE (unit, "(' : ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%step)
    END IF
    
  END SUBROUTINE foralltriplet_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallTripletSpec%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'forall-header'
  !!
  !! We may specialize this for the extensions of Expression at a later 
  !! date.
  
  FUNCTION foralltriplet_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'forall-triplet-spec'
    
  END FUNCTION foralltriplet_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallTripletSpec%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE foralltriplet_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallTripletSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%start%Query(visitor)
    
    CALL pnode%finish%Query(visitor)
    
    IF (ALLOCATED(pnode%step)) THEN
      CALL pnode%step%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE foralltriplet_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallTripletSpec%Modify.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE foralltriplet_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallTripletSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%start%Modify(visitor)
    
    CALL pnode%finish%Modify(visitor)
    
    IF (ALLOCATED(pnode%step)) THEN
      CALL pnode%step%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE foralltriplet_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallTripletSpec%GetLocation - get the start of 
  !! the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION foralltriplet_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ForallTripletSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .Locationof. pnode%index
    
  END FUNCTION foralltriplet_GetLocation
  
END MODULE ForallStmts
