! $Id: IfStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IfStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /if-then-stmt/'s, /if-stmt/'s, /else-if-stmt/'s, 
!! /else-stmt/'s, /end-if-stmt/'s and /arithmetic-if-stmt/'s.
!!
!! Also procedures and types for parsing if-constructs.

MODULE IfStmts
  
  USE Expressions
  USE Labels
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIfStmt
  PUBLIC :: ParseIfThenStmt
  PUBLIC :: ParseElseIfStmt
  PUBLIC :: ParseElseStmt
  PUBLIC :: ParseEndIfStmt
  PUBLIC :: ParseArithmeticIfStmt
  
  PUBLIC :: Expression        ! From the Expressions module for convenience.
  
  !> Stores the results of parsing an arithmetic if statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ArithmeticIfStmt
    !> The expression tested by the statement.
    CLASS(Expression), ALLOCATABLE :: expr
    !> Labels that follow the statement.
    TYPE(Label) :: the_labels(3)
  CONTAINS
    PROCEDURE :: Dump => arith_Dump
    PROCEDURE :: Query => arith_Query
    PROCEDURE :: Modify => arith_Modify
    PROCEDURE, NOPASS :: What => arith_What
  END TYPE ArithmeticIfStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /if-stmt/ (R837).
  !!
  !! This is the single line form of if statement.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /if-stmt/.
  !!
  !! @param[out]    expr              The expression that is tested by the 
  !! if-stmt.
  !!
  !! @param[out]    child_tlist       List of tokens that make up the child 
  !! /action-stmt/ of the /if-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIfStmt(tlist, expr, child_tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Token), INTENT(OUT), ALLOCATABLE :: child_tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Component for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R837'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /if-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! IF ...
    
    CALL ExpectKeyword(tlist, 1, scck_'IF', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! IF ( /expr/ ) ...
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, 3, closing - 1, expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! IF ( /expr/ ) action-stmt
    
    IF (SIZE(tlist) == closing) THEN
      ! The error message is cognisant of classification options.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(closing)),  &
          MSG='Expected /action-stmt/, "THEN" or label after IF (...).' )
      RETURN
    END IF
    
    child_tlist = tlist(closing+1:)
    
  END SUBROUTINE ParseIfStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /if-then-stmt/ (R833).  
  !! This is the construct form of if statement.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /if-stmt/.
  !!
  !! @param[out]    expr              The expression that is tested by the 
  !! /if-stmt/.
  !!
  !! @param[out]    name              The /construct-name/ in the 
  !! /if-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIfThenStmt(tlist, expr, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /if-then-stmt/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R833'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    INTEGER :: closing        ! End index of parenthesis after expression
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /if-then-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Check for construct name
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    ! IF ....
    CALL ExpectKeyword(tlist, it, scck_'IF', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! IF (...
    CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, it + 2, closing - 1, expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! IF (xxx) THEN
    CALL ExpectKeyword(tlist, closing + 1, scck_'THEN', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Thats it!
    CALL ExpectEOS(tlist, closing + 2, comp, err_list)
    
  END SUBROUTINE ParseIfThenStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-if-stmt/ (R836).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-if-stmt/.
  !!
  !! @param[in]    name               The /construct-name/ in the opening 
  !! /if-stmt/ if one was present, not allocated if a construct name was 
  !! not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndIfStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-if-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R836'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-if-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDIF or END IF ...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'IF', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... /construct-name/
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        name,  &
        '/if-then-stmt/',  &
        '/end-if-stmt/',  &
        'C827 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndIfStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /else-stmt/ (R835).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /else-stmt/.
  !!
  !! @param[in]     name              The /construct-name/ in the opening 
  !! /if-stmt/ if one was present, not allocated if a construct name was 
  !! not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseElseStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /else-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R835'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /else-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ELSE ...
    
    CALL ExpectKeyword(tlist, 1, scck_'ELSE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional construct name
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        2,  &
        name,  &
        '/if-then-stmt/',  &
        '/else-stmt/',  &
        'C827 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseElseStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /else-if-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /else-if-stmt/.
  !!
  !! @param[in]    name               The /construct-name/ in the opening 
  !! /if-stmt/ if one was present, not allocated if a construct name was 
  !! not present.
  !!
  !! @param[out]    expr              The expression that is tested by the 
  !! /else-if-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseElseIfStmt(tlist, construct_name, expr, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /else-if-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R834'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /else-if-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ELSEIF or ELSE IF ...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'ELSE', scck_'IF', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! it here is the index of the thing after IF.
    
    !---------------------------------------------------------------------------
    ! ... ( expr )
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, it + 1, closing - 1, expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! ... THEN
    
    CALL ExpectKeyword(tlist, closing + 1, scck_'THEN', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... construct-name
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        closing + 2,  &
        construct_name,  &
        '/if-then-stmt',  &
        '/else-if-stmt/',  &
        'C827 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseElseIfStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /arithmetic-if-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /arithmetic-if-stmt/.
  !!
  !! @param[out]    arith_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This is a deprecated language feature.
  
  SUBROUTINE ParseArithmeticIfStmt(tlist, arith_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ArithmeticIfStmt), INTENT(OUT) :: arith_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /arithmetic-if-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R853'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /arithmetic-if-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! IF ...
    
    CALL ExpectKeyword(tlist, 1, scck_'IF', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    it = 1
    
    !---------------------------------------------------------------------------
    ! ... ( expr )
    
    CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, it + 2, closing - 1, arith_stmt%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Skip trailing parenthesis.
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! ... /label/, /label/, /label/
    
    ! Note there must be three!
    
    ! Label one.
    CALL expect_label(tlist, it, arith_stmt%the_labels(1), sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    it = it + 1
    
    ! Comma one.
    CALL Expect(tlist, it, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=COMP)
    IF (Failure(err)) RETURN
    it = it + 1
    
    ! Label two.
    CALL expect_label(tlist, it, arith_stmt%the_labels(2), sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    it = it + 1
    
    ! Comma two.
    CALL Expect(tlist, it, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=COMP)
    IF (Failure(err)) RETURN
    it = it + 1
    
    ! Label three.
    CALL expect_label(tlist, it, arith_stmt%the_labels(3), sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    it = it + 1
    
    ! That's it.
    CALL ExpectEOS(tlist, it, comp, err_list)
    
  END SUBROUTINE ParseArithmeticIfStmt
  
  
  !*****************************************************************************
  !!
  !> Expect and parse a label at a particular token.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     pos               Index where the label is expected.
  !!
  !! @param[out]    label             The value of the label.  Not defined 
  !! in the event of error.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE expect_label(tlist, pos, the_label, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: pos
    TYPE(Label), INTENT(OUT) :: the_label
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Child error list for ParseLabel call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (pos > SIZE(tlist)) THEN
      CALL Add(err_list,  &
          CODE=errSyntax,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /label/.' )
      RETURN
    END IF
    
    CALL ParseLabel(  &
        tlist,  &
        pos,  &
        pos,  &
        the_label,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE expect_label
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArithmeticIfStmt%Dump - write a debugging 
  !! representation of the results of parsing an arithmetic if statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE arith_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArithmeticIfStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    ! The expression.
    WRITE (unit, "(A,'/scalar-numeric-expr/ = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, "()")
    ! The child labels list.
    WRITE (unit, "(A,'labels = ',I0,', ',I0,', ',I0)") pfx,  &
        .ValueOf. pnode%the_labels
    
  END SUBROUTINE arith_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArithmeticIfStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE arith_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArithmeticIfStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE arith_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArithmeticIfStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE arith_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArithmeticIfStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE arith_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArithmeticIfStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'arithmetic-if-stmt'.
  
  FUNCTION arith_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'arithmetic-if-stmt'
    
  END FUNCTION arith_What
  
END MODULE IfStmts
