! $Id: InputOutputItems.f90 2800 2019-03-22 18:13:05Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the InputOutputItems module.


!*******************************************************************************
!!
!> Type and associated procedures for storing input/output items.
!!
!! R916 /input-item/    is /variable/
!!                      or /io-implied-do/
!!
!! R917 /output-item/   is /expr/
!!                      or /io-implied-do/
!!
!! The difference between /expr/'s and /variable/'s is carried through into 
!! the permissable objects in the /io-implied-do/.
!!
!! We currently don't parse variables with ParseVariable, or store the 
!! results in the Variable type.  To do so would require splitting the 
!! procedures in this module to some extent.

MODULE InputOutputItems
  
  USE CompilerKinds
  USE Expressions
  USE ImpliedDos
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: ParseIOItemList
  PUBLIC :: QueryIOItemList
  PUBLIC :: ModifyIOItemList
  PUBLIC :: DumpIOItemList
  
  !-----------------------------------------------------------------------------
  
  !> Parent class for the various things that might be in an /input-item-list/ 
  !! or an /output-item-list/.
  !!
  !! We us SCParseNode as the parent, because io lists can get a bit 
  !! lengthy and confusing, so we want the ability to Dump to multiple 
  !! lines.
  TYPE, EXTENDS(SCParseNode), PUBLIC, ABSTRACT :: IOItem
  END TYPE IOItem
  
  !> An element for a vector of IOItems.
  TYPE, PUBLIC :: IOItemList
    CLASS(IOItem), ALLOCATABLE :: Item
  END TYPE IOItemList
  
  !> An output-item that is an expression.
  TYPE, PUBLIC, EXTENDS(IOItem) :: IOItemExpr
    !> The output item is an expression.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => ioitem_Dump
    PROCEDURE :: Query => ioitem_Query
    PROCEDURE :: Modify => ioitem_Modify
    PROCEDURE, NOPASS :: What => ioitem_What
  END TYPE IOItemExpr
  
  !> Represents an /io-implied-do-control/.
  TYPE, PUBLIC, EXTENDS(ImpliedDoControl) :: IOImpliedDoControl
  CONTAINS
    PROCEDURE, NOPASS :: What => ioimpldocontrol_What
    PROCEDURE :: Advertise => ioimpldocontrol_Advertise
  END TYPE IOImpliedDoControl
  
  !> An input or output item that is an /io-implied-do/.
  TYPE, PUBLIC, EXTENDS(IOItem) :: IOItemImpliedDo
    !> The output item is an implied do.
    TYPE(IOItemList), ALLOCATABLE :: items(:)
    !> Implied do control.
    TYPE(IOImpliedDoControl) :: control
  CONTAINS
    PROCEDURE :: Dump => ioimpldo_Dump
    PROCEDURE :: Query => ioimpldo_Query
    PROCEDURE :: Modify => ioimpldo_Modify
    PROCEDURE, NOPASS :: What => ioimpldo_What
  END TYPE IOItemImpliedDo
  
  !-----------------------------------------------------------------------------
  
  ! Syntax rule number for an /output-item/.
  CHARACTER(*), PARAMETER :: out_comp = 'R917'
  
  ! Syntax rule number for an /input-item/.
  CHARACTER(*), PARAMETER :: in_comp = 'R916'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /input-item-list/ or an 
  !! /output-item-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the item list.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the item list.
  !!
  !! @param[in]     output            Indicates that an /output-item-list/ 
  !! is being parsed.
  !!
  !! @param[out]    items             The results of parsing the list.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIOItemList(tlist, start, finish, output, items, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    LOGICAL, INTENT(IN) :: output
    TYPE(IOItemList), ALLOCATABLE, INTENT(OUT) :: items(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the list.
    INTEGER :: item_start     ! Index of the starting token for an item.
    INTEGER :: item_finish    ! Index of the finishing token for an item.
    INTEGER :: item           ! Item index.
    
    ! Error list for child parsing procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=MERGE(out_comp, in_comp, output),  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /' // rule(output) // 'put-item-list/.')
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(start:finish))
    ALLOCATE(items(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      
      CALL parse_io_item(  &
          tlist,  &
          item_start, item_finish,  &
          output,  &
          items(item)%item,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseIOItemList
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /input-item/ (R916) or 
  !! /output-item/ (R917).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the item.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the item.
  !!
  !! @param[in]     output            Indicates that an /output-item-list/ 
  !! is being parsed.
  !!
  !! @param[out]    item              The results of parsing the item.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! An IOItem may be an impied do, which itself contains IOItems, hence 
  ! this is recursive.
  RECURSIVE SUBROUTINE parse_io_item( tlist, start, finish, output,  &
      item, err_list )
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    LOGICAL, INTENT(IN) :: output
    CLASS(IOItem), INTENT(OUT), ALLOCATABLE :: item
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of closing parenthesis.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: assign         ! Index of the assign (=) token.
    
    ! Object to use if the item is an io-implied-do.
    CLASS(IOItemImpliedDo), ALLOCATABLE :: implied_do_item
    
    ! Object to use if the item is an expression.
    CLASS(IOItemExpr), ALLOCATABLE :: expr_item
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=MERGE(out_comp, in_comp, output),  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /' // rule(output) // 'put-item/.')
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Do we have an IO implied do?  
    !
    ! The form is (list, var=start, finish [, step]) 
    !
    ! Is the first token a parenthesis?
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      ! Find the end of the parenthesised bit.
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=MERGE(out_comp, in_comp, output))
      IF (Failure(err)) RETURN
      
      assign = FindBare(tlist(start+1:closing-1), ittAssign, scck_'=')
      IF (assign /= 0) THEN
        ! We appear to have an /ac-implied-do/ as the item
        ALLOCATE(implied_do_item)
        
        CALL parse_implied_do_content(  &
            tlist,  &
            start + 1, closing - 1,  &
            output,  &
            implied_do_item,  &
            sub_err_list )
        CALL Add( err_list,  &
            sub_err_list,  &
            COMPONENT=MERGE(out_comp, in_comp, output) )
        IF (Failure(sub_err_list)) RETURN
        
        CALL MOVE_ALLOC(implied_do_item, item)
        
        CALL ExpectEnd(  &
            tlist(:finish),  &
            closing + 1,  &
            MERGE(out_comp, in_comp, output),  &
            err_list )
        RETURN
      END IF
    END IF
    
    ! The item is an expression.
    ALLOCATE(expr_item)
    CALL ParseExpr(tlist, start, finish, expr_item%expr, sub_err_list)
    CALL Add( err_list,  &
        sub_err_list,  &
        COMPONENT=MERGE(out_comp, in_comp, output) )
    IF (Failure(sub_err_list)) RETURN
    
    CALL MOVE_ALLOC(expr_item, item)
  
  END SUBROUTINE parse_io_item
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /io-implied-do/ (R918).
  !!
  !! @param[in]     tlist             The sequence of tokens.  Should not 
  !! include the starting and ending parentheses.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! after the opening parenthesis of the implied do.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! before the closing parenthesis of the implied do.
  !!
  !! @param[in]     output            Indicates that an /output-item-list/ 
  !! is being parsed.
  !!
  !! @param[out]    implied_do        Results of parsing the /io-implied-do/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The list items can be /input-item/'s or /output-item/'s - we don't 
  !! distinguish here.
  
  SUBROUTINE parse_implied_do_content( tlist, start, finish, output,  &
      implied_do, err_list )
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    LOGICAL, INTENT(IN) :: output
    TYPE(IOItemImpliedDo), INTENT(OUT) :: implied_do
    TYPE(Error), ALLOCATABLE, INTENT(OUT) :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /io-implied-do/.
    CHARACTER(*), PARAMETER :: comp = 'R918'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of value items in an /io-implied-do/.
    INTEGER :: item_start     ! Starting index of an item.
    INTEGER :: item_finish    ! Finishing index of an item.
    INTEGER :: item           ! Item index.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    CALL ParseImpliedDo(  &
        tlist,  &
        start,  &
        finish,  &
        'io-implied-do-object',  &
        '/io-implied-do/',  &
        comp,  &
        list_count,  &
        implied_do%control,  &
        err_list )
    
    ALLOCATE(implied_do%items(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start,  item_finish)
      CALL parse_io_item(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          output,  &
          implied_do%items(item)%item,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END DO
    
  END SUBROUTINE parse_implied_do_content
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for IOItemExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemExpr%Dump - write a text representation of the 
  !! node to a logical unit.
  !!
  !! @param[in]     pnode             The node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  
  SUBROUTINE ioitem_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOItemExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL pnode%expr%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE ioitem_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemExpr%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE ioitem_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOItemExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ioitem_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemExpr%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE ioitem_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOItemExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ioitem_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemExpr%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION ioitem_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    ! either an /input-item/ or an /output-item/ - depends on context.
    what = '(io-item)'
    
  END FUNCTION ioitem_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for IOItemImpliedDo.
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemImpliedDo%Dump - dump a text representation of 
  !! the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  
  SUBROUTINE ioimpldo_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOItemImpliedDo), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    CALL DumpIOItemList(pnode%items, unit, pfx // '  ')
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL pnode%control%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE ioimpldo_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemImpliedDo%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE ioimpldo_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOItemImpliedDo), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL QueryIOItemList(pnode%items, visitor)
    
    CALL pnode%control%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ioimpldo_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemImpliedDo%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE ioimpldo_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOItemImpliedDo), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL ModifyIOItemList(pnode%items, visitor)
    
    CALL pnode%control%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ioimpldo_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOItemImpliedDo%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION ioimpldo_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    ! either an /input-item/ or an /output-item/ - depends on context.
    what = 'io-implied-do'
    
  END FUNCTION ioimpldo_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for IOImpliedDoControl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOImpliedDoControl%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION ioimpldocontrol_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'io-implied-do-control'
    
  END FUNCTION ioimpldocontrol_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOImpliedDoControl%Advertise - advertise 
  !! the identifiers referenced or defined by the array value.
  !!
  !! Unlike the "default" implementation of Advertise for implied do control, 
  !! a io-implied-do does not start a new scope and hence does not 
  !! declare a new entity.
  !!
  !! @param[in]     pnode             The implied do control to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  !!
  !! The do variable is given a special definition nature, because it may 
  !! be a construct local variable.
  
  ! Recursive because an implied do control can contain another implied do 
  ! control if the programmer is being ridiculous.
  RECURSIVE SUBROUTINE ioimpldocontrol_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOImpliedDoControl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%do_variable,  &
        NATURES=[NatureRefDataObject],  &
        LOCATIONS=[.LocationOf. pnode%do_variable] )
    
  END SUBROUTINE ioimpldocontrol_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for IOItemList.
  
  
  !*****************************************************************************
  !!
  !> Forwarder for the Query call to an IOItemList object.
  !!
  !! @param[in]     list_item         The IOItemList object.
  !!
  !! @param[in,out] visitor           The visitor object executing the query 
  !! operation.
  
  ! Recursive because item lists can contain item lists.
  RECURSIVE SUBROUTINE QueryIOItemList(list, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IOItemList), INTENT(IN) :: list(:)
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! List index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(list)
      CALL list(i)%item%Query(visitor)
    END DO
    
  END SUBROUTINE QueryIOItemList
  
  
  !*****************************************************************************
  !!
  !> Forwarder for the Modify call to an IOItemList object.
  !!
  !! @param[in]     list_item         The IOItemList object.
  !!
  !! @param[in,out] visitor           The visitor object executing the modify 
  !! operation.
  
  ! Recursive because item lists can contain item lists.
  RECURSIVE SUBROUTINE ModifyIOItemList(list, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IOItemList), INTENT(INOUT) :: list(:)
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! List index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(list)
      CALL list(i)%item%Modify(visitor)
    END DO
    
  END SUBROUTINE ModifyIOItemList
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of an OutputItemList to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     items             The OutputItemList to dump.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  
  ! Recursive because item lists can contain item lists.
  RECURSIVE SUBROUTINE DumpIOItemList(list, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IOItemList), INTENT(IN) :: list(:)
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! List index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(list)
      CALL list(i)%item%Dump(unit, pfx)
    END DO
    
  END SUBROUTINE DumpIOItemList
  
  
  !*****************************************************************************
  !!
  !> Return the prefix of the item list syntax rule name.
  !!
  !! @param[in]     output            Indicates the relevant rule is the 
  !! output rule.
  !!
  !! @returns 'out' if @a output is true, otherwise 'in'.
  
  PURE FUNCTION rule(output)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    LOGICAL, INTENT(IN) :: output
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: rule
    
    !***************************************************************************
    
    IF (output) THEN
      rule = 'out'
    ELSE
      rule = 'in'
    END IF
    
  END FUNCTION rule
  
END MODULE InputOutputItems
