! $Id: InterfaceStmts.f90 2924 2019-04-26 23:28:12Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the InterfaceStmts module.


!*******************************************************************************
!!
!> Procedure for parsing /interface-stmt/'s, /end-interface-stmt/'s and 
!! /procedure-stmt/'s (because they only appear inside interface blocks.

MODULE InterfaceStmts
  
  USE GenericSpecs
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseInterfaceStmt
  PUBLIC :: ParseEndInterfaceStmt
  PUBLIC :: ParseProcedureStmt
  
  PUBLIC :: GenericSpec       ! From GenericSpecs
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse an /interface-stmt/ (R1203).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    is_abstract       If true the statement is the 
  !! opening statement of an ABSTRACT interface block.
  !!
  !! @param[out]    generic_spec      If allocated this is the 
  !! generic specification that appeared after the INTERFACE keyword.  Not 
  !! allocated if there was no such specification.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /continue-stmt/ consists solely of the keyword "CONTINUE".
  
  SUBROUTINE ParseInterfaceStmt(tlist, is_abstract, generic_spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE GenericSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(OUT) :: is_abstract
    TYPE(GenericSpec), INTENT(OUT), ALLOCATABLE :: generic_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /interface-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1203'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /interface-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    !  [ ABSTRACT ] INTERFACE
    
    CALL ExpectKeyword(  &
        tlist,  &
        1,  &
        ['ABSTRACT ', 'INTERFACE'],  &
        '"ABSTRACT INTERFACE" or "INTERFACE"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      is_abstract = .TRUE.
      
      ! ABSTRACT INTERFACE is all that the statement can be.
      CALL ExpectKeyword( tlist, 2, scck_'INTERFACE', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = 3
      
      CALL ExpectEOS(tlist, 3, comp, err_list)
    ELSE
      is_abstract = .FALSE.
      
      ! The generic spec is optional.
      IF (2 > SIZE(tlist)) RETURN
      
      ALLOCATE(generic_spec)
      CALL ParseGenericSpec(tlist(2:), generic_spec, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
  END SUBROUTINE ParseInterfaceStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that supposedly represent an 
  !! /end-interface-stmt/ (R1204).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     generic_spec      The generic spec on the opening 
  !! /interface-stmt/ if present, not allocated if none was present.
  !!
  !! @param[out]    err_list          List of errors.
    
  SUBROUTINE ParseEndInterfaceStmt(tlist, generic_spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE GenericSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(GenericSpec), INTENT(IN), ALLOCATABLE :: generic_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /end-interface-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1204'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object from Expect etc.
    INTEGER :: after          ! Position after the END TYPE keywords. 
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Generic spec on this statement.
    TYPE(GenericSpec) :: end_generic
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-interface-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Either END INTERFACE or ENDINTERFACE
    
    CALL ExpectSplittableKeyword(  &
        tlist,  &
        1,  &
        scck_'END',  &
        scck_'INTERFACE',  &
        after,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The generic spec is optional
    IF (after > SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ParseGenericSpec(tlist(after:), end_generic, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! Check the provided generic spec matches what we are looking for.  
    ! Fall through for the error message.
    IF (ALLOCATED(generic_spec)) THEN
      IF (end_generic%keyword == generic_spec%keyword) THEN
        SELECT CASE (generic_spec%keyword)
        CASE (0)    ! A generic name
          IF (.NameOf. end_generic%name == .NameOf. generic_spec%name) RETURN
        CASE (1)    ! An operator
          IF (is_same_operator(  &
                .NameOf. end_generic%name,  &
                .NameOf. generic_spec%name ))  &
              RETURN
        CASE (2:)    ! ASSIGNMENT(=), READ(FORMATTED)
          RETURN
        END SELECT
      END IF
    END IF
    
    CALL Add( err_list,  &
        CODE=errNameMismatch,  &
        COMPONENT='C1202 on ' // comp,  & 
        LOCATION=QueryLocation(tlist(after)),  &
        MSG='If a /generic-spec/ is provided in an /end-interface-stmt/ &
          &it shall shall match the /generic-spec/ in the corresponding &
          &/interface-stmt/.' )
    
  END SUBROUTINE ParseEndInterfaceStmt
  
  
  !*****************************************************************************
  !!
  !> Test whether two operator tokens refer to the same operator.
  !!
  !! @param[in]     lhs               One of the operator tokens.
  !!
  !! @param[in]     rhs               The other operator token.
  !!
  !! @returns True if the operator tokens are the same (.EQ. is the same 
  !! as ==, for example), false otherwise.
  
  FUNCTION is_same_operator(lhs, rhs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CHARACTER(*), INTENT(IN) :: lhs
    CHARACTER(*), INTENT(IN) :: rhs
    
    ! Function result
    LOGICAL b
    
    !***************************************************************************
    
    IF ((lhs == '.EQ.') .OR. (lhs == '==')) THEN
      b = (rhs == '.EQ.') .OR. (rhs == '==')
      RETURN
    END IF
    
    IF ((lhs == '.NE.') .OR. (lhs == '/=')) THEN
      b = (rhs == '.NE.') .OR. (rhs == '/=')
      RETURN
    END IF
    
    IF ((lhs == '.GE.') .OR. (lhs == '>=')) THEN
      b = (rhs == '.GE.') .OR. (rhs == '>=')
      RETURN
    END IF
    
    IF ((lhs == '.LE.') .OR. (lhs == '<=')) THEN
      b = (rhs == '.LE.') .OR. (rhs == '<=')
      RETURN
    END IF
    
    IF ((lhs == '.GT.') .OR. (lhs == '>')) THEN
      b = (rhs == '.GT.') .OR. (rhs == '>')
      RETURN
    END IF
    
    IF ((lhs == '.LT.') .OR. (lhs == '<')) THEN
      b = (rhs == '.LT.') .OR. (rhs == '<')
      RETURN
    END IF
    
    b = lhs == rhs
    
  END FUNCTION is_same_operator
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that supposedly represent an 
  !! /procedure-stmt/ (R1206).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    has_module        True if the statement has the module 
  !! qualifier, otherwise false.
  !!
  !! @param[in]     procedure_names   The names of the produres.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! If the objective is to just extract the procedure name without the 
  !! full parse of the statement, then use 
  !! ProcedureStmtUtils::GetProcedureName.
    
  SUBROUTINE ParseProcedureStmt(tlist, has_module, procedure_names, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE GenericSpecs
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(OUT) :: has_module
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: procedure_names(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /procedure-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1206'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object from Expect etc.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /procedure-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Either MODULE PROCEDURE or PROCEDURE
    
    CALL Expect(  &
        tlist,  &
        1,  &
        [ittName, ittName],  &
        [scck_'MODULE   ', scck_'PROCEDURE'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)      ! MODULE
      has_module = .TRUE.
      ! Now need PROCEDURE
      CALL Expect(tlist, 2, ittName, scck_'PROCEDURE', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = 3
      
    CASE (2)      ! PROCEDURE
      has_module = .FALSE.
      it = 2
      
    END SELECT
    
    ! it here index after the PROCEDURE keyword.
    
    !---------------------------------------------------------------------------
    ! Optional double colon.
    
    CALL Expect(  &
        tlist, it,  &
        [ittName, ittDeclare],  &
        [scck_'  ', scck_'::'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Skip double colon if present.
    IF (opt == 2) it = it + 1
    
    !---------------------------------------------------------------------------
    ! The /procedure-name-list/.
    
    CALL ParseNameList(  &
        tlist,  &
        it,  &
        SIZE(tlist),  &
        '/procedure-name/',  &
        procedure_names,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! All done!  The name list parse above picks up any extraneous tokens.
    
  END SUBROUTINE ParseProcedureStmt
  
END MODULE InterfaceStmts
