! $Id: Labels.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Labels module.


!*******************************************************************************
!!
!> Defines a type that can act as common storage for labels.
!!
!! There are similarities between this type and Identifier, but this is 
!! a little lower level.
!!
!! Because errors may be generated when parsing a label from a token 
!! (for example - a kind part is not valid) we don't have defined assignment 
!! from a token to a label.

MODULE Labels
  
  USE LabelsUtils
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseLabel
  PUBLIC :: Label             ! From LabelsUtils.
  PUBLIC :: MaxLabelLength    ! From LabelsUtils.
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens (one token really) that is a statement label.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the label.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the label.
  !!
  !! @param[out]    the_label         The results of parsing the label.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An error message of some sort will ultimately be generated if @a start is 
  !! not the same as @a finish - a label is always a single token.
  
  SUBROUTINE ParseLabel(tlist, start, finish, the_label, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Label), INTENT(OUT) :: the_label
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child parsing and scanning calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=label_comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /label/.' )
      RETURN
    END IF
    
    ! We deliberately don't look at the type of the token, as the syntax 
    ! rules define label as its own token type, even though it is 
    ! indistinguishable from a kindless integer.
    CALL ScanLabel(  &
        QueryRaw(tlist(start)),  &
        QueryLocation(tlist(start)),  &
        the_label,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    IF (Failure(sub_err_list)) RETURN
    
    CALL ExpectEnd(tlist(:finish), start + 1, label_comp, err_list)
    
  END SUBROUTINE ParseLabel
  
END MODULE Labels
