! $Id: NamelistStmts.f90 2922 2019-04-26 22:40:09Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the NamelistStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /namelist-stmt/'s (R563).
!!
!! Why does something that was added to the language relatively recently 
!! (F90) have such archaic syntax?

MODULE NamelistStmts
  
  USE CompilerKinds
  USE Identifiers
  USE ParseNodes
  USE Strings
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseNamelistStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a namelist name and object list combination.
  !!
  !! There's no format definition of a namelist group - it is a construct 
  !! that we've invented ourselves to simplify parsing.
  TYPE, PUBLIC :: NamelistGroup
    !> The name of the namelist group.
    TYPE(Identifier) :: name
    !> The variables in the namelist group.
    TYPE(Identifier), ALLOCATABLE :: objects(:)
  END TYPE NamelistGroup
  
  !> Stores the results of parsing a namelist statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: NamelistStmt
    !> The results of parsing the namelist groups defined by the statement.
    !!
    !! Always allocated and SIZE > 1 after successful construction.
    TYPE(NamelistGroup), ALLOCATABLE :: groups(:)
  CONTAINS
    PROCEDURE :: Dump => nml_Dump
    PROCEDURE :: Query => nml_Query
    PROCEDURE :: Modify => nml_Modify
    PROCEDURE, NOPASS :: What => nml_What
  END TYPE NamelistStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /namelist-stmt/ (R563).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    nml_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseNamelistStmt(tlist, nml_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(NamelistStmt), INTENT(OUT) :: nml_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /namelist-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R563'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Index of the current token.
    INTEGER :: it_end         ! End of the current syntax part.
    INTEGER :: opt            ! Expect option.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Temporary for growing the list of sets.
    TYPE(NamelistGroup), ALLOCATABLE :: tmp_groups(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /namelist-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! NAMELIST...
    
    CALL ExpectKeyword(tlist, 1, scck_'NAMELIST', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The namelist groups.
    
    ALLOCATE(nml_stmt%groups(1))
    it = 2
    DO
      ! Progressive parse - it_end gets set to the token one past the end 
      ! of the namelist group.
      CALL parse_namelist_group(  &
          tlist,  &
          it,  &
          nml_stmt%groups(SIZE(nml_stmt%groups)),  &
          it_end,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (it_end > SIZE(tlist)) EXIT
      it = it_end
      
      ! Create space for the next item
      ALLOCATE(tmp_groups(SIZE(nml_stmt%groups) + 1))
      tmp_groups(:SIZE(nml_stmt%groups)) = nml_stmt%groups
      
      CALL MOVE_ALLOC(tmp_groups, nml_stmt%groups)
      
      ! A comma is optional.
      ! Failure not covered by StmtParsingTests.
      CALL Expect(  &
          tlist, it,  &
          [ittComma, ittOperator],  &
          [scck_',', scck_'/'],  &
          '"," or "/"',  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Move over the comma if necessary
      IF (opt == 1) it = it + 1
    END DO
    
  END SUBROUTINE ParseNamelistStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a namelist group.
  !!
  !! We define a namelist group to be:
  !!
  !! / /namelist-group-name/ / /namelist-group-object-list/
  !!
  !! Note there is no trailing slash after the object list.
  !!
  !! Because this is our own definition we don't chain a syntax rule number 
  !! to any error messages.
  !!
  !! @param[in]     tlist             The sequence of tokens for the 
  !! statement.
  !!
  !! @param[in]     start             The index of the start of the namelist 
  !! group in @a tlist.
  !!
  !! @param[out]    group             Results of parsing the namelist group.  
  !! Not defined on error.
  !!
  !! @param[out]    after             The index of the token in @a tlist that 
  !! immediately follows the namelist group.  This does not include the 
  !! optional separating comma, just to be consistenent with our definition 
  !! of a namelist group above.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE parse_namelist_group(tlist, start, group, after, err_list)
    
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(NamelistGroup), INTENT(OUT) :: group
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: slash          ! Index of the slash after the object list.
    INTEGER :: list_finish    ! Index of the end of the object list.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! A slash precedes the name.
    
    CALL Expect(tlist, start, ittOperator, scck_'/', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /namelist-group-name/.
    
    CALL ExpectName(tlist, start + 1, '/namelist-group-name/', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    group%name = IdentifierIfortBug(tlist(start+1))
    
    !---------------------------------------------------------------------------
    ! Another slash
    
    CALL Expect(tlist, start + 2, ittOperator, scck_'/', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! A /namelist-group-object-list/.  This is terminated either by a slash, 
    ! or by a comma immediately followed by a slash, or by the end of the 
    ! statement.
    
    ! Find the /.  If we don't find it we process the remainder of the 
    ! statement as the list.
    slash = FindBare(tlist, ittOperator, scck_'/', start + 3)
    IF (slash == 0) THEN
      ! If this is the end of the statement (no slash to delineate the next 
      ! group) then a trailing comma is not allowed.  If one is there we will 
      ! error out in the processing of the object list.
      list_finish = SIZE(tlist)
    ELSE
      ! Do we have a comma immediately preceding the terminating slash?
      IF (tlist(slash-1) == ittComma) THEN
        list_finish = slash - 2
      ELSE
        ! No trailing comma.
        list_finish = slash - 1
      END IF
    END IF
    
    ! The object list itself
    CALL ParseNameList(  &
        tlist,  &
        start + 3, list_finish,  &
        '/namelist-group-object/',  &
        group%objects,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! finish here is the end of the list of objects.  Bump it to be one past 
    ! the end.  This may be the comma.
    after = list_finish + 1
    
  END SUBROUTINE parse_namelist_group
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamelistStmt%Dump - write a debugging representation 
  !! of the results of parsing an intent statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE nml_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamelistStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ig             ! Group index.
    INTEGER :: io             ! Object index.
    
    !***************************************************************************
    
    DO ig = 1, SIZE(pnode%groups)
      WRITE (unit, "(A,'/namelist-group-name/ = ',A)") pfx,  &
          .NameOf. pnode%groups(ig)%name
      DO io = 1, SIZE(pnode%groups(ig)%objects)
        WRITE (unit, "(A,A)")  &
            pfx // '  ',  &
            .NameOf. pnode%groups(ig)%objects(io)
      END DO
    END DO
    
  END SUBROUTINE nml_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamelistStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE nml_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamelistStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nml_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamelistStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE nml_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamelistStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nml_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamelistStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'namelist-stmt'.
  
  FUNCTION nml_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'namelist-stmt'
    
  END FUNCTION nml_What
  
END MODULE NamelistStmts
