! $Id: ParseImplicitPartStmtListExternal.f90 2908 2019-04-13 00:28:11Z ian $
! ff08 source code copyright 2019 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParseImplicitPartStmtListExternal procedure.


!*******************************************************************************
!!
!> Worker routine for ParseSpecificationPart to handle the implicit_part_nodes 
!! component of the statement tree node for the specification part.
!!
!! @param[in]     stmts             The array of statement tree nodes for the 
!! implicit part.  All the nodes must be valid in an implicit part.
!!
!! @param[out]    children          The results of parsing the array of 
!! statement tree nodes.
!!
!! @param[in,out] err_list          List of errors.

RECURSIVE SUBROUTINE ParseImplicitPartStmtListExternal( stmts,  &
    children, err_list )
  
  USE Errors
  USE ParseNodes
  USE StNodes
  
  !-----------------------------------------------------------------------------
  ! Arguments
  
  TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
  TYPE(SCParseNodeElement), INTENT(INOUT), ALLOCATABLE :: children(:)
  TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
  !***************************************************************************
  
  CALL execute(stmts, children, err_list)
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> The main execution sequence for the subroutine.  Split out to avoid 
  !! accidental re-use of host associated names.
  !!
  !! Arguments have the same meaning as for the host.
  
  RECURSIVE SUBROUTINE execute(stmts, children, err_list)
    
    USE Statements
    USE SingleStNodes
    
    USE EntryStmts
    USE FormatStmts
    USE ImplicitStmts
    USE ParameterStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
    TYPE(SCParseNodeElement), INTENT(INOUT), ALLOCATABLE :: children(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! Statement tree node index.
    
    ! Generic parse node temporary.
    CLASS(SCParseNode), ALLOCATABLE :: pnode
    
    ! ParseNode for an entry statement.
    TYPE(EntryStmt), ALLOCATABLE :: entry_stmt
    
    ! ParseNode for a format statement.
    TYPE(FormatStmt), ALLOCATABLE :: fmt_stmt
    
    ! ParseNode for an implicit statement.
    TYPE(ImplicitStmt), ALLOCATABLE :: impl_stmt
    
    ! ParseNode for a parameter statement.
    TYPE(ParameterStmt), ALLOCATABLE :: param_stmt
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    DO i = 1, SIZE(stmts)
      ! Similar pattern to other factory procedures - allocate a local object 
      ! of the relevant ParseNode extension, use a ParseXxx procedure to 
      ! construct it, on success move that extension object into a polymorphic 
      ! pnode object, on failure deallocate the extension object.  After 
      ! the select block we add the polymorphic object to the array of 
      ! child objects.
      SELECT TYPE (item => stmts(i)%item)
      TYPE IS (SingleStNode)
        SELECT CASE (item%stmt%ist)
        CASE (istEntry)
          ALLOCATE(entry_stmt)
          CALL ParseEntryStmt(item%stmt%tlist, entry_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(entry_stmt)
          ELSE
            CALL MOVE_ALLOC(entry_stmt, pnode)
          END IF
        
        CASE (istFormat)
          ALLOCATE(fmt_stmt)
          CALL ParseFormatStmt(item%stmt%tlist, fmt_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(fmt_stmt)
          ELSE
            CALL MOVE_ALLOC(fmt_stmt, pnode)
          END IF
        
        CASE (istImplicit)
          ALLOCATE(impl_stmt)
          CALL ParseImplicitStmt(item%stmt%tlist, impl_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(impl_stmt)
          ELSE
            CALL MOVE_ALLOC(impl_stmt, pnode)
          END IF
        
        CASE (istParameter)
          ALLOCATE(param_stmt)
          CALL ParseParameterStmt(item%stmt%tlist, param_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(param_stmt)
          ELSE
            CALL MOVE_ALLOC(param_stmt, pnode)
          END IF
        
        CASE DEFAULT
          STOP 'Internal error in &
              &SpecificationParts%%process_implicit_part_statements, &
              &encountered an unknown type of statement.'
        
        END SELECT
      CLASS DEFAULT
        PRINT "('The unknown statement tree node is for a ',A)",  &
            GetPartName(item)
        STOP 'In SpecificationParts%%process_implicit_part_statements, &
            &unknown StNode.'
      END SELECT
      CALL Add(err_list, sub_err_list)
      IF (.NOT. Failure(sub_err_list)) THEN
        CALL pnode%Init(stmts(i)%item)
        CALL AddPNode(children, pnode)
      END IF
    END DO
    
  END SUBROUTINE execute
  
END SUBROUTINE ParseImplicitPartStmtListExternal
