! $Id: ParseNodes.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParseNodes module.


!*******************************************************************************
!!
!> Defines the ParseNode type, that acts as a common parent type for 
!! all statement and construct parsing results.
!!
!! No matter what I do, everything turns into some sort of tree data structure.
!!
!! In only makes sense to use this type as the parent for statement parsing 
!! results where the statement stands alone in a parent construct.  For 
!! example, there's no point parsing the opening if-then statement in an 
!! if construct to something that is an extension of a parse node, because 
!! the if-then statement is always part of an if construct.
!!
!! This currently simply specifies an interface for debug output, nothing 
!! more.  In the near term and extension of the ParseNode type may need to 
!! be created for constructs that have a construct name, with appropriate 
!! bindings and means of navigating the hierarchy of parse nodes in order 
!! to relate construct names to child statements.
!!
!! Similarly, we might need to track statement labels - which might be 
!! better done at the ParseNode level (though statements don't know what 
!! their own label is, as the token is split off in tokenization).
!!
!! To facilitate later significant changes in design, we have started using 
!! the interface of the ParseNode type for any types that might store the 
!! results of parsing, even if those types are below statement level.  If 
!! we do go down the path of forcing a common inheritance, we'll probably 
!! need to reorganise ParseNode into ConstructParseNode or similar.

MODULE ParseNodes
  
  USE Errors
  USE Identifiers
  USE LabelsUtils
  USE SourceLocations
  USE Strings
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: AddPNode
  PUBLIC :: Check
  PUBLIC :: ParseNameList
  
  !-----------------------------------------------------------------------------
  
  !> Parent type for the results of parsing all statements and constructs, 
  !! where those results may need to be stored.
  TYPE, PUBLIC, ABSTRACT :: ParseNode
  CONTAINS
    PROCEDURE(pnode_What), DEFERRED, NOPASS :: What
    
    ! Tree traversal.
    ! @todo Extensions should have been required to implement a doQuery 
    ! and doModify binding that could be private, with the public 
    ! bindings (or just public procedures) providing the 
    ! wrapper calls to start and finish and then invoking that 
    ! private binding in between.  One day.
    PROCEDURE(pnode_Query), DEFERRED :: Query
    PROCEDURE(pnode_Modify), DEFERRED :: Modify
    
    ! Location information.
    PROCEDURE(pnode_GetLocation), DEFERRED :: GetLocation
    GENERIC :: OPERATOR(.LocationOf.) => GetLocation
    
    ! Operations on a ParseNode element.
    PROCEDURE :: Advertise => pnode_Advertise
    PROCEDURE :: Check => pnode_Check
  END TYPE ParseNode
  
  !> A visitor to a node for a query operation in the parse tree.
  TYPE, PUBLIC, ABSTRACT :: ParseNodeQueryVisitor
  CONTAINS
    PROCEDURE :: Start => pnodequeryvisitor_Start
    PROCEDURE :: Finish => pnodequeryvisitor_Finish
  END TYPE ParseNodeQueryVisitor
  
  !> A visitor to a node for a modify operation in the parse tree.
  TYPE, PUBLIC, ABSTRACT :: ParseNodeModifyVisitor
  CONTAINS
    PROCEDURE :: Start => pnodemodifyvisitor_Start
    PROCEDURE :: Finish => pnodemodifyvisitor_Finish
  END TYPE ParseNodeModifyVisitor
  
  ! Interfaces for the deferred bindings in ParseNode.
  ABSTRACT INTERFACE
    
    !> Interface for ParseNode%What - return a descriptive string 
    !! for the parse node.
    !!
    !! @returns A string that describes what the node is.  This would normally 
    !! by a syntax rule name, without the leading and trailing slashes.
    !!
    !! For whatever reason, we've decided to return a string rather than the 
    !! syntax rule number as used in StNode equivalent.  The string is more 
    !! flexible, but might have been a mistake.
    FUNCTION pnode_What() RESULT(what)
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      ! Function result.
      CHARACTER(:), ALLOCATABLE :: what
    END FUNCTION pnode_What
    
    !> Interface for ParseNode%Query - traverse the parse node with a 
    !! query operation.
    !!
    !! @param[in]     pnode           The parse node being queried.
    !!
    !! @param[in,out] visitor         The object carrying out the operation.
    !!
    !! Parse nodes that have child nodes should invoke the Start binding 
    !! of the @a visitor on themselves, then pass the query call onto each 
    !! of their children, then invoke the Finish binding.
    SUBROUTINE pnode_Query(pnode, visitor)
      IMPORT :: ParseNode
      IMPORT :: ParseNodeQueryVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(ParseNode), INTENT(IN) :: pnode
      CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE pnode_Query
    
    !> Interface for ParseNode%Modify - traverse the parse node with a 
    !! modify operation.
    !!
    !! @param[in]     pnode           The parse node being modified.
    !!
    !! @param[in,out] visitor         The object carrying out the operation.
    !!
    !! Parse nodes that have child nodes should invoke the Start binding 
    !! of the @a visitor on themselves, then pass the modify call onto each 
    !! of their children, then invoke the Finish binding of the visitor.
    SUBROUTINE pnode_Modify(pnode, visitor)
      IMPORT :: ParseNode
      IMPORT :: ParseNodeModifyVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(ParseNode), INTENT(INOUT) :: pnode
      CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE pnode_Modify
    
    !> Interface of ParseNode%GetLocation - get the start of the source 
    !! that corresponds to the ParseNode.
    !!
    !! @param[in]     pnode           The parse node of interest
    !!
    !! @returns The location in source of the start of the ParseNode.
    
    FUNCTION pnode_GetLocation(pnode) RESULT(loc)
      USE SourceLocations
      IMPORT :: ParseNode
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(ParseNode), INTENT(IN) :: pnode
      ! Function result.
      TYPE(SourceLocation) :: loc
    END FUNCTION pnode_GetLocation
  END INTERFACE
  
  !-----------------------------------------------------------------------------
  
  !> A parse node to use for parse nodes that represet statements or 
  !! constructs.
  !!
  !! Both of these sorts of ParseNodes support the Dump operation.
  TYPE, ABSTRACT, EXTENDS(ParseNode), PUBLIC :: SCParseNode
    !> A statement label associated with the node.  Not allocated if there 
    !!
    TYPE(Label), ALLOCATABLE :: statement_label
    !> The location of the start of the parse node.
    !!
    !! This is set in InitParseNode.  This is becoming a little pointless as 
    !! time goes on, because the nodes have often also got a stored location 
    !! in the parsed data that they are storing.
    TYPE(SourceLocation), PRIVATE :: location
  CONTAINS
    PROCEDURE(scpn_Dump), DEFERRED :: Dump
    PROCEDURE :: GetLocation => scpn_GetLocation
    
    PROCEDURE :: InitStNode => scpn_InitStNode
    PROCEDURE :: InitStData => scpn_InitStData
    GENERIC :: Init => InitStNode, InitStData
  END TYPE SCParseNode
  
  ABSTRACT INTERFACE
    !> Interface for ParseNode%Dump - write a debugging representation 
    !! of the parse node to a logical unit.
    !!
    !! @param[in]     pnode             The ParseNode to dump.
    !!
    !! @param[in]     unit              Logical unit connected for 
    !! formatted output.
    !!
    !! @param[in]     pfx               Prefix for each record.
    SUBROUTINE scpn_Dump(pnode, unit, pfx)
      IMPORT :: SCParseNode
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(SCParseNode), INTENT(IN) :: pnode
      INTEGER, INTENT(IN) :: unit
      CHARACTER(*), INTENT(IN) :: pfx
    END SUBROUTINE scpn_Dump
  END INTERFACE
  
  !> An element in an array of SCParseNodes.
  !!
  !! We never expect this to be extended, but we provide some bindings that 
  !! simply forward the call to the item component for convenience.
  !!
  !! (This has a very similar interface to the SCParseNode type itself, but it 
  !! cannot be an extension of ParseNode because the What binding is NOPASS, 
  !! meaning we would not be able to access the item component to forward 
  !! the call.)
  TYPE, PUBLIC :: SCParseNodeElement
    !> The item in the array.
    CLASS(SCParseNode), ALLOCATABLE :: item
  CONTAINS
    ! These simply forward to the item component.
    PROCEDURE :: Dump => scpne_Dump
    PROCEDURE :: Query => scpne_Query
    PROCEDURE :: Modify => scpne_Modify
    PROCEDURE :: What => scpne_What
  END TYPE SCParseNodeElement
  
  !-----------------------------------------------------------------------------
  
  !> A parse node to use for parse nodes that represent terms (things smaller 
  !! than statements or constructs).
  TYPE, ABSTRACT, EXTENDS(ParseNode), PUBLIC :: TermParseNode
  CONTAINS
    PROCEDURE(tpn_Write), DEFERRED :: Write
  END TYPE TermParseNode
  
  ABSTRACT INTERFACE
    SUBROUTINE tpn_Write(pnode, unit)
      IMPORT :: TermParseNode
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(TermParseNode), INTENT(IN) :: pnode
      INTEGER, INTENT(IN) :: unit
    END SUBROUTINE tpn_Write
  END INTERFACE
  
  !-----------------------------------------------------------------------------
  
  !> Common ParseNode for statements that just involve a list of names.
  TYPE, ABSTRACT, EXTENDS(SCParseNode), PUBLIC :: NameListPNode
    !> The names in the statement.
    TYPE(Identifier), ALLOCATABLE :: name_list(:)
  CONTAINS
    PROCEDURE :: Dump => nlpnode_Dump
    PROCEDURE :: Query => nlpnode_Query
    PROCEDURE :: Modify => nlpnode_Modify
    
    PROCEDURE(nlpnode_GetListName), DEFERRED, NOPASS :: GetListName
  END TYPE NameListPNode
  
  ! Interfaces for the deferred bindings in NameListPNode.
  ABSTRACT INTERFACE
    !> Interface for NameListPNode%GetListName - return a name for the 
    !! stored list of names.
    !!
    !! @returns A name for use in the implementation of the Dump binding 
    !! to report the stored names.
    FUNCTION nlpnode_GetListName() RESULT(name)
      IMPLICIT NONE
      CHARACTER(:), ALLOCATABLE :: name
    END FUNCTION nlpnode_GetListName
  END INTERFACE
  
  !-----------------------------------------------------------------------------
  
  !> Run Check on a parse tree.
  INTERFACE Check
    MODULE PROCEDURE Check_
  END INTERFACE Check
  
  !> Type to execute the check traversal.
  TYPE, EXTENDS(ParseNodeQueryVisitor) :: check_query
    !> The progressively assembled error list.
    TYPE(Error), ALLOCATABLE :: err_list(:)
  CONTAINS
    PROCEDURE :: Finish => check_Finish
  END TYPE check_query
  
CONTAINS
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ParseNode
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNode%Check - check the parse node for conformance 
  !! with language standards.
  !!
  !! @param[in]     pnode           The parse node of interest.
  !!
  !! @param[out]    err_list        List of errors.
  !!
  !! The default implementation checks nothing.
  
  SUBROUTINE pnode_Check(pnode, err_list)
    
    USE Errors
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNode), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
  END SUBROUTINE pnode_Check
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNode%Advertise - advertise the identifiers 
  !! referenced or defined by the node (excluding child nodes), and the 
  !! nature of that reference or definition.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        The identifiers defined or referenced 
  !! by the node.
  
  SUBROUTINE pnode_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNode), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE pnode_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ParseNodeQueryVisitor.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNodeQueryVisitor%Start - called to execute the 
  !! action of the visitor against a particular parse node before 
  !! any child nodes are processed.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     pnode             The node to execute the action against.
  !!
  !! The default implementation does nothing.
  
  SUBROUTINE pnodequeryvisitor_Start(visitor, pnode)
    
    !-------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(IN) :: pnode
    
    !***************************************************************************
    
    ! Does nothing.
    
  END SUBROUTINE pnodequeryvisitor_Start
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNodeQueryVisitor%Start - called to execute the 
  !! action of the visitor against a particular parse node after 
  !! all child nodes are processed.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     pnode             The node to execute the action against.
  !!
  !! The default implementation does nothing.
  
  SUBROUTINE pnodequeryvisitor_Finish(visitor, pnode)
    
    !-------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(IN) :: pnode
    
    !***************************************************************************
    
    ! Does nothing.
    
  END SUBROUTINE pnodequeryvisitor_Finish
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ParseNodeModifyVisitor.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNodeModifyVisitor%Start - called to execute the 
  !! action of the visitor against a particular parse node before 
  !! any child nodes are processed.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     pnode             The node to execute the action against.
  !!
  !! The default implementation does nothing.
  
  SUBROUTINE pnodemodifyvisitor_Start(visitor, pnode)
    
    !-------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(INOUT) :: pnode
    
    !***************************************************************************
    
    ! Does nothing.
    
  END SUBROUTINE pnodemodifyvisitor_Start
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParseNodeModifyVisitor%Start - called to execute the 
  !! action of the visitor against a particular parse node after 
  !! all child nodes are processed.
  !!
  !! @param[in,out] visitor           The visitor object.
  !!
  !! @param[in]     pnode             The node to execute the action against.
  !!
  !! The default implementation does nothing.
  
  SUBROUTINE pnodemodifyvisitor_Finish(visitor, pnode)
    
    !-------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(INOUT) :: pnode
    
    !***************************************************************************
    
    ! Does nothing.
    
  END SUBROUTINE pnodemodifyvisitor_Finish
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for SCParseNodeElement.
  
  
  !*****************************************************************************
  !!
  !> Move a child item into an array of parse nodes.
  !!
  !! @param[in,out] children          The array of parse nodes.
  !!
  !! @param[in,out] item              The item to add.  This will be moved 
  !! across to the list (the actual argument will be unallocated after 
  !! this procedure exits.
  
  SUBROUTINE AddPNode(children, item)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SCParseNodeElement), INTENT(INOUT), ALLOCATABLE :: children(:)
    CLASS(SCParseNode), INTENT(INOUT), ALLOCATABLE :: item
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for growing array.
    TYPE(SCParseNodeElement), ALLOCATABLE :: tmp(:)
    
    INTEGER :: i              ! Array index.
    
    !***************************************************************************
    
    IF (ALLOCATED(children)) THEN
      ALLOCATE(tmp(LBOUND(children,1):UBOUND(children,1)+1))
      DO i = LBOUND(children, 1), UBOUND(children, 1)
        CALL MOVE_ALLOC(children(i)%item, tmp(i)%item)
      END DO
    ELSE
      ALLOCATE(tmp(1))
    END IF
    
    CALL MOVE_ALLOC(item, tmp(UBOUND(tmp,1))%item)
    CALL MOVE_ALLOC(tmp, children)
    
  END SUBROUTINE AddPNode
  
  
  !*****************************************************************************
  !!
  !> Initially define (construct) the common parts of a ParseNode from the 
  !! corresponding statement tree node.
  !!
  !! @param[in,out] pnode             The parse node being initially defined.
  !!
  !! @param[in]     st                The corresponding statement tree node.
  !!
  !! The ParseNode%Init binding must be called as part of the construction 
  !! of a ParseNode.
  
  SUBROUTINE scpn_InitStNode(pnode, st)
    
    USE StNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SCParseNode), INTENT(INOUT) :: pnode
    CLASS(StNode), INTENT(IN) :: st
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Temporary to allow us to extract the GetLabel call result.
    TYPE(LabelWrapper) :: label_tmp
    
    !***************************************************************************
    
    ! Save the starting location of the corresponding statement tree node 
    ! as the starting location of the parse node.
    pnode%location = st%GetLocation()
    
    ! Save the label of the statement tree node.
    label_tmp = st%GetLabel()
    CALL MOVE_ALLOC(label_tmp%item, pnode%statement_label)
    
  END SUBROUTINE scpn_InitStNode
  
  
  !*****************************************************************************
  !!
  !> Initially define (construct) the common parts of a ParseNode from the 
  !! corresponding statement tree node.
  !!
  !! @param[in,out] pnode             The parse node being initially defined.
  !!
  !! @param[in]     st                The corresponding statement tree node.
  !!
  !! The ParseNode%Init binding must be called as part of the construction 
  !! of a ParseNode.
  
  SUBROUTINE scpn_InitStData(pnode, stmt)
    
    USE StatementData
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SCParseNode), INTENT(INOUT) :: pnode
    TYPE(StData), INTENT(IN) :: stmt
    
    !***************************************************************************
    
    ! Save the starting location of the corresponding statement tree node 
    ! as the starting location of the parse node.
    pnode%location = QueryLocation(stmt%tlist(1))
    
    ! Save the label of the statement tree node.
    IF (ALLOCATED(stmt%statement_label)) THEN
      pnode%statement_label = stmt%statement_label
    END IF
    
  END SUBROUTINE scpn_InitStData
  
  
  !*****************************************************************************
  !!
  !> Implementation of SCParseNodeElement%Dump - forward the Dump call to the 
  !! item.
  !!
  !! @param[in]     pnode             The parse node element to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE scpne_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SCParseNodeElement), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    CALL pnode%item%Dump(unit, pfx)
    
  END SUBROUTINE scpne_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SCParseNodeElement%Query - forward the Query call to the 
  !! item.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE scpne_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SCParseNodeElement), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Query(visitor)
    
  END SUBROUTINE scpne_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SCParseNodeElement%Mod=ify - forward the Modify call 
  !! to the item.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE scpne_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SCParseNodeElement), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Modify(visitor)
    
  END SUBROUTINE scpne_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SCParseNodeElement%What - forward the What call to the 
  !! item.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns The result of calling the What binding of the item.
  
  FUNCTION scpne_What(pnode) RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SCParseNodeElement), INTENT(IN) :: pnode
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = pnode%item%What()
    
  END FUNCTION scpne_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SCParseNode%GetLocation - get the start of the source 
  !! that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION scpn_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SCParseNode), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION scpn_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for NameListPNode.
  
  
  !*****************************************************************************
  !!
  !> Implementation of NameListPNode%Dump - write a debugging representation 
  !! of the results of parsing an external statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE nlpnode_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NameListPNode), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! List index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,A)") pfx, pnode%GetListName()
    DO i = 1, SIZE(pnode%name_list)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%name_list(i)
    END DO
    
  END SUBROUTINE nlpnode_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of NameListPNode%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  !!
  !! This default implementation simply forwards the call to the Start 
  !! and Finish bindings of the visitor.
  
  ! Recursive in case a query executes another query.
  RECURSIVE SUBROUTINE nlpnode_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(NameListPNode), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! Nothing to do but visit us.
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nlpnode_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of NameListPNode%Modify.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  !!
  !! This default implementation simply forwards the call to the Start 
  !! and Finish bindings of the visitor.
  
  ! Recursive in case a query executes another query.
  RECURSIVE SUBROUTINE nlpnode_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(NameListPNode), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! Nothing to do but visit us.
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nlpnode_Modify
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Support for the Check sub-pass.
  
  
  !*****************************************************************************
  !!
  !> Run the check sub-pass against a parse tree.
  !!
  !! @param[in]     pnode             The root of the tree.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE Check_(pnode, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNode), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Object to execute the check traversal.
    TYPE(check_query) :: visitor
    
    !***************************************************************************
    
    ALLOCATE(visitor%err_list(0))
    
    CALL pnode%Query(visitor)
    
    CALL MOVE_ALLOC(visitor%err_list, err_list)
    
  END SUBROUTINE Check_
  
  
  !*****************************************************************************
  !!
  !> Implementation of check_query%Execute - executes the Check action on 
  !! a particular node after its children have been Checked.
  !!
  !! @param[in,out] visitor           The object executing the traversal.
  !!
  !! @param[in]     pnode             The current node being visited.
  
  SUBROUTINE check_Finish(visitor, pnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(check_query), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(IN) :: pnode
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for this Check call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    CALL pnode%Check(sub_err_list)
    CALL Add(visitor%err_list, sub_err_list)
    
  END SUBROUTINE check_Finish
  
END MODULE ParseNodes
