! $Id: ParserTest.f90 2930 2019-04-27 23:22:26Z ian $
!> @file
!! Defines the ParserTest program.


!*****************************************************************************
!!
!> Test driver for the Parse phase of constructing the syntax tree.
!!
!! Just takes a single argument - the name of the file to parse.
!!
!! Exit status: 
!! - 0: if no errors were detected
!! - 1: if parsing errors were detected.
!! - 3: you botched the command line.

PROGRAM ParserTest
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Locals
  
  ! Format specification for an error message.
  CHARACTER(*), PARAMETER :: fmt_error = "(A)"
  
  !*****************************************************************************
  
  CALL run
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Execute the parse test (internal procedure to avoid issues with host 
  !! association.
  
  SUBROUTINE run
    
    USE BaseStNodes
    USE BuildTreePass
    USE CharUtils
    USE CmdLine
    USE DefaultHosts
    USE Errors
    USE ParseNodes
    USE Programs
    USE Strings
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: OUTPUT_UNIT, ERROR_UNIT
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Command line options.
    !
    ! In order:
    !   --checks
    !   --console
    TYPE(CmdLineOption) :: cmd_line_options(2)
    
    ! List of non-option arguments.
    TYPE(String), ALLOCATABLE :: arguments(:)
    
    ! Error message
    CHARACTER(:), ALLOCATABLE :: msg
    
    INTEGER :: stat             ! Command line retrieval status code.
    
    ! The root of the statement tree for the file.
    TYPE(BaseStNode), POINTER :: base
    
    ! List of errors from various procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    INTEGER :: ie               ! Error index
    INTEGER :: stop_code        ! Stop code.
    
    ! The root of the parse tree for the file.
    TYPE(Program) :: program_node
    
    ! The source form chosen for the console, or all blanks if the console 
    ! is not being read.
    CHARACTER(:), ALLOCATABLE :: console
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Parse the command line and get the filename.
    
    cmd_line_options(1) = CmdLineOption(  &
        'checks',  &
        DESCRIPTION='Carry out additional language &
          &related checks on the tree.' )
    
    cmd_line_options(2) = CmdLineOption(  &
        'console',  &
        DESCRIPTION='Reads input source from the console, after any other &
          &other files.  The argument specifies the source form, if not &
          &specified the default is `free` for Fortran 90 &
          &free form source.',  &
        OPTION_TYPE=ioptArgOpt )
    
    CALL ParseCmdLine(cmd_line_options, arguments, stat, msg)
    
    IF (stat /= 0) THEN
      WRITE (ERROR_UNIT, "('Error parsing command line:',A)") msg
      STOP 3
    END IF
    
    IF (cmd_line_options(2)%PresentCount /= 0) THEN
      IF (SIZE(cmd_line_options(2)%Argument) /= 0) THEN
        console = cmd_line_options(2)%Argument(1)%item
        CALL UpperCase(console)
      ELSE
        console = 'FREE'
      END IF
    ELSE
      console = ''
    END IF
    
    !---------------------------------------------------------------------------
    ! Build the statement tree.
    
    CALL build_combined_statement_tree(  &
        arguments,  &
        console,  &
        ERROR_UNIT,  &
        base,  &
        stop_code )
    IF (stop_code /= 0) CALL do_stop(stop_code)
    
    CALL DoBuildTreePassCheck(base, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (ERROR_UNIT, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) STOP 1
    
    !---------------------------------------------------------------------------
    ! Parse the statement tree.
    
    CALL ParseProgram(base, program_node, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (ERROR_UNIT, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) STOP 1
    
    !---------------------------------------------------------------------------
    ! Carry out additional checks, if requested.
    
    IF (cmd_line_options(1)%PresentCount /= 0) THEN
      CALL Check(program_node, err_list)
      DO ie = 1, SIZE(err_list)
        WRITE (ERROR_UNIT, fmt_error) .Describe. err_list(ie)
      END DO
      IF (Failure(err_list)) STOP 1
    END IF
    
    !---------------------------------------------------------------------------
    ! Dump the parsed represenation of the program.
    
    CALL program_node%Dump(OUTPUT_UNIT, '  ')
    
  END SUBROUTINE run
  
  
  !*****************************************************************************
  !!
  !> Build the combined statement tree for all files.
  
  SUBROUTINE build_combined_statement_tree( filenames, console, err_unit,  &
      combined, stop_code )
    
    USE BaseStNodes
    USE BuildTreePass
    USE DefaultHosts
    USE ScanFixedForm
    USE ScanFreeForm
    USE Sources
    USE Scanner
    USE SourceUnits
    USE Strings
    USE Errors
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: INPUT_UNIT
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(String), INTENT(IN) :: filenames(:)
    CHARACTER(*), INTENT(IN) :: console
    INTEGER, INTENT(IN) :: err_unit
    TYPE(BaseStNode), INTENT(OUT), POINTER :: combined
    INTEGER, INTENT(OUT) :: stop_code
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ie             ! Error index.
    INTEGER :: if             ! File index.
    
    ! Compiler host object to define host behaviour.
    TYPE(DefaultHost) :: host
    
    ! Error object for procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    ! The statement tree for the current file.
    TYPE(BaseStNode), POINTER :: current
    
    ! Source form for reading from the console.
    CLASS(SourceForm), ALLOCATABLE :: console_form
    
    ! Source object connected to the input unit.
    TYPE(SourceUnit), POINTER :: source_unit
    
    ! Pointer to source_unit.
    CLASS(Source), POINTER :: p_source
    
    !***************************************************************************
    
    ALLOCATE(combined)
    
    DO if = 1, SIZE(filenames)
      CALL DoBuildTreePass(  &
          host,  &
          filenames(if)%item,  &
          current,  &
          err_list )
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_error) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END DO
    
    ! If enabled, read statements from the console.
    IF (console /= '') THEN
      CALL ConnectSourceUnit(INPUT_UNIT, source_unit)
      p_source => source_unit
      SELECT CASE (console)
      CASE ('FIXED')
        ALLOCATE(console_form, SOURCE=FixedFormSource(p_source))
      CASE ('FREE')
        ALLOCATE(console_form, SOURCE=FreeFormSource(p_source))
      END SELECT
      
      CALL DoBuildTreePass(  &
          host,  &
          console_form,  &
          current,  &
          err_list)
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_error) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END IF
    
    stop_code = 0
    
  END SUBROUTINE build_combined_statement_tree
  
  
  !*****************************************************************************
  !!
  !> Map an integer stop code across to a stop statement.
  !!
  !! @param[in]     stop_code         Stop code for the program.
  !!
  !! This is a work-around for the inability in Fortran 2003 to use an 
  !! expression (or a non-constant expression in Fortran 2008) as the 
  !! stop code in a stop statement.
  
  SUBROUTINE do_stop(stop_code)
    INTEGER, INTENT(IN) :: stop_code
    
    SELECT CASE (stop_code)
    CASE (1)  ; STOP 1
    CASE (2)  ; STOP 2
    CASE (3)  ; STOP 3
    CASE (4)  ; STOP 4
    END SELECT
  END SUBROUTINE do_stop
  
END PROGRAM ParserTest
