! $Id: PointerAssignmentStmts.f90 2902 2019-04-11 21:03:41Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PointerAssignmentStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /pointer-assignment-stmt/'s (R733).
!!
!! The left hand side can be a /data-pointer-object/ or a 
!! /proc-pointer-object/.  
!! - A /data-pointer-object/ looks like (is?) a /variable/, but 
!!   with constraints that mean it looks like a designator.
!! - A /proc-pointer-object/ similarly looks ultimately like a 
!!   constrained form of designator.
!!
!! The right hand side can be either a /data-target/ or a /proc-target/, 
!! respectively.
!! - A /data-target/ is a variable, which we parse as an expression.
!! - A /proc-target/ similarly is parsed as an expression.

MODULE PointerAssignmentStmts
  
  USE Designators
  USE Expressions
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParsePointerAssignmentStmt
  
  ! Exports for the convenience of users.
  PUBLIC :: PartRef           ! From the Designators module.
  PUBLIC :: Expression        ! From the Expressions module.
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a pointer assignment statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: PointerAssignmentStmt
    !> The designator for the pointer.
    !!
    !! Always allocated, size greater than zero.
    TYPE(PartRef), ALLOCATABLE :: pointer(:)
    !> The target for the assignment.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: target
  CONTAINS
    PROCEDURE :: Dump => ptrassign_Dump
    PROCEDURE :: Query => ptrassign_Query
    PROCEDURE :: Modify => ptrassign_Modify
    PROCEDURE, NOPASS :: What => ptrassign_What
  END TYPE
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /pointer-assignment-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ptrassign_stmt    The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParsePointerAssignmentStmt(tlist, ptrassign_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(PointerAssignmentStmt), INTENT(OUT) :: ptrassign_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /pointer-assignment-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R733'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: assign         ! Index of the "=>" token.
    INTEGER :: percent        ! Index of the "%" token.
    INTEGER :: it             ! Current token index.
    INTEGER :: ipt            ! Part index.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /pointer-assignment-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Find the location of the '=>'
    assign = FindBare(tlist, ittAssociate, scck_'=>')
    IF (assign == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='This /pointer-assignment-stmt/ appears to be missing a "=>".' )
      RETURN
    ELSE IF (assign == 1) THEN
      ! Trap this so we don't get a bizarre "expected expression" error 
      ! from the expression parsing code.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='Expected /data-pointer-object/ or /proc-pointer-object/.' )
      RETURN
    END IF
    
    ! Parse the pointer bit.
    ALLOCATE(ptrassign_stmt%pointer(CountParts(tlist, 1, assign - 1)))
    it = 1
    DO ipt = 1, SIZE(ptrassign_stmt%pointer)
      percent = FindBare(tlist(:assign-1), ittPercent, scck_'%', it)
      IF (percent == 0) percent = assign
      CALL ParsePartRef(  &
          tlist,  &
          it, percent - 1,  &
          ptrassign_stmt%pointer(ipt),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = percent + 1
    END DO
    
    ! Parse the expression bit.
    CALL ParseExpr(  &
        tlist,  &
        assign + 1, SIZE(tlist),  &
        ptrassign_stmt%target,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParsePointerAssignmentStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerAssignmentStmt%Dump - write a debugging 
  !! representation of the results of parsing a pointer assignment statement 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE ptrassign_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerAssignmentStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'pointer = ')", ADVANCE='NO') pfx
    DO i = 1, SIZE(pnode%pointer)
      CALL WritePartRef(unit, pnode%pointer(i))
      WRITE (unit, "(' % ')", ADVANCE='NO')
    END DO
    WRITE (unit, "()")
      
    WRITE (unit, "(A,'target = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%target)
    WRITE (unit, "()")
    
  END SUBROUTINE ptrassign_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptrassign_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerAssignmentStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Miscelleous index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%pointer)
      CALL pnode%pointer(i)%Query(visitor)
    END DO
    
    CALL pnode%target%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptrassign_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptrassign_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerAssignmentStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Miscelleous index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%pointer)
      CALL pnode%pointer(i)%Modify(visitor)
    END DO
    
    CALL pnode%target%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptrassign_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerAssignmentStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'pointer-assignment-stmt'.
  
  FUNCTION ptrassign_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'pointer-assignment-stmt'
    
  END FUNCTION ptrassign_What
  
END MODULE PointerAssignmentStmts
