! $Id: ProtectedStmts.f90 2196 2016-06-07 02:56:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ProtectedStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /protected-stmt/'s.

MODULE ProtectedStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseProtectedStmt
  
  TYPE, EXTENDS(NameListPNode), PUBLIC :: ProtectedStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => prot_GetListName
    PROCEDURE, NOPASS :: What => prot_What
    PROCEDURE :: Advertise => prot_Advertise
  END TYPE ProtectedStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /protected-stmt/ (R552).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    prot_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseProtectedStmt(tlist, prot_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProtectedStmt), INTENT(OUT) :: prot_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /protected-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R552'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /intent-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! PROTECTED...
    
    CALL ExpectKeyword(tlist, 1, scck_'PROTECTED', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /dummy-arg-name/
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /dummy-arg-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /dummy-arg-name-list/ here.
    
    !---------------------------------------------------------------------------
    ! The /dummy-arg-name-list.
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/dummy-arg-name/',  &
        prot_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseProtectedStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProtectedStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/dummy-arg-name-list/'
  
  FUNCTION prot_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/dummy-arg-name-list/'
    
  END FUNCTION prot_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExternalStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'external-stmt'.
  
  FUNCTION prot_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'protected-stmt'
    
  END FUNCTION prot_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProtectedStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE prot_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProtectedStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[NatureDefProtected], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE prot_Advertise
  
END MODULE ProtectedStmts
