! $Id: ReadStmts.f90 2896 2019-04-07 01:14:52Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ReadStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /read-stmt/'s (R904).

MODULE ReadStmts
  
  USE Formats, ONLY: Format
  USE InputOutputItems
  USE IoControlSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseReadStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /read-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ReadStmt
    !> The /io-control-spec-list/.
    !!
    !! Allocated only if the read statement is of the `READ (xxx)` form.
    TYPE(IoControlSpecList), ALLOCATABLE :: io_control_spec_list
    !> The format specification of the /read-stmt/.
    !!
    !! Allocated only if the read statement is of the `READ fmt` form.
    TYPE(Format), ALLOCATABLE :: fmt
    !> The list of input items.
    !!
    !! Always allocated, but might be zero size.
    TYPE(IOItemList), ALLOCATABLE :: item_list(:)
  CONTAINS
    PROCEDURE :: Dump => read_Dump
    PROCEDURE :: Query => read_Query
    PROCEDURE :: Modify => read_Modify
    PROCEDURE, NOPASS :: What => read_What
    PROCEDURE :: Check => read_Check
  END TYPE ReadStmt
  
  !-----------------------------------------------------------------------------
  ! Local constants
  
  ! Syntax rule number for a /read-stmt/.
  CHARACTER(*), PARAMETER :: comp = 'R910'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /read-stmt/ (R910).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    read_stmt         The results of parsing the /read-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseReadStmt(tlist, read_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ReadStmt), INTENT(OUT) :: read_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /read-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    !READ...
    
    CALL ExpectKeyword(tlist, 1, scck_'READ', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Trickiness trying to work out what form of read statement we have.
    
    ! .TRUE. for first form, .FALSE. otherwise.
    IF (test_read_statement_form(tlist)) THEN
      ! Definitely first form
      CALL parse_first_form(tlist, read_stmt, err_list)
    ELSE
      ! Most likely second form
      CALL parse_second_form(tlist, read_stmt, err_list)
    END IF
    
  END SUBROUTINE ParseReadStmt
  
  
  !*****************************************************************************
  !!
  !> Test the form of a /read-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that makes up the 
  !! /read-stmt/.
  !!
  !! @returns .TRUE. if the token sequence appears to describe the first form 
  !! of /read-stmt/, .FALSE. otherwise.
  !!
  !! The first form is the classic:
  !!
  !!   READ ( /io-spec-control-list/ ) [ /input-item-list/ ]
  !!
  !! The second form is the PRINT analogue.
  !!
  !!   READ /format/ [ , /input-item-list/ ]
  !!
  !! As the first form is more common, we will default to it where syntax 
  !! errors may be obscuring the programmers intent.
  !!
  !! - If the second token is not a "(", then it can only validly be the 
  !!   second form.
  !!
  !! - If there is a bare "," in the parenthesised list, then it can only 
  !!   validly be the first form.
  !!
  !! - If there are any argument keywords inside the parenthesised list, then 
  !!   it can only validly be the first form.
  !!
  !! - If there is a comma following the closing parenthesis, then it can 
  !!   only validly be the second form.
  !!
  !! F2008 permits variables to be function references (in the general sense).
  !! That means that is possible for a unary operator that follows the closing 
  !! parenthesis to be part of a function reference for the first variable 
  !! in the input item list.  At this stage of parsing we don't know whether 
  !! an operator is a suitable unary operator (even if it is one of the 
  !! "standard" symbol like operators, as they could be overloaded if the 
  !! next operand is of derived type).
  !!
  !! Furthermore, even if we did have all the details about the operator, it 
  !! is possible to construct syntax ambiguous forms by having an operator 
  !! that has both unary (with a data pointer result) and binary (with a 
  !! character result that is a valid format specification) forms.
  !!
  !! Consequently, we will also assume:
  !!
  !! - If there is any operator following the closing parenthesis, then it 
  !!   will be assumed to be the second form.
  !!
  !! and let later code dealing with the second form decide (if it can) that 
  !! a parsing mistake was made.
  
  FUNCTION test_read_statement_form(tlist) RESULT(first_form)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    LOGICAL :: first_form
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: finish         ! End of the parenthesised bit.
    
    !***************************************************************************    
    
    ! Watch for "double negatives" in the following logic.
  
    ! Second token a "(" ?
    first_form = Match(tlist, 2, ittDelimiter, scck_'(')
    IF (.NOT. first_form) RETURN   ! No - can only validly be second form.
    
    ! Token two is a "(" from here on, first_form "defaulted" by above to 
    ! .TRUE..  Get the closing parenthesis location.
    closing = FindClosingParens(tlist, 2)
    IF (closing == 0) THEN
      ! Tolerate a mismatch and see if we can deduce the programmers intent.
      finish = SIZE(tlist)
    ELSE
      finish = closing - 1
    END IF
    
    ! Bare "," inside the parenthesis?
    first_form = FindBare(tlist(3:finish), ittComma, scck_',') /= 0
    IF (first_form) RETURN    ! Can only validly be first form.
    
    ! Argument keywords?
    first_form = FindBare(tlist(3:finish), ittAssign, scck_'=') /= 0
    IF (first_form) RETURN    ! Can only validly be first form.
    
    ! If we have mismatched parentheses (an error) and no comma or 
    ! argument style keyword to let us know what's intended, then assume 
    ! this is the classic first form for now and let the parsing code deal 
    ! with the syntax error.
    first_form = (closing == 0)
    IF (first_form) RETURN    ! No closing parenthesis - take as first form.
    
    ! No comma following closing parenthesis?
    first_form = .NOT. Match(tlist, closing + 1, ittComma, scck_',')
    IF (.NOT. first_form) RETURN   ! Can only validly be second form.

    ! Entering the grey zone... see comments in procedure preamble.
    
    ! No operator following closing parenthesis?
    first_form = .NOT. Match(tlist, closing + 1, ittOperator)
    IF (.NOT. first_form) RETURN
    
    first_form = .TRUE.
    
  END FUNCTION test_read_statement_form
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the first form of /read-stmt/.
  !!
  !! This is the form that is the analogue of a /write-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    read_stmt         The results of parsing the /read-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_first_form(tlist, read_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ReadStmt), INTENT(OUT) :: read_stmt
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ! Failure not tested by StmtParsingTests as this is a basic requirement 
    ! of the first form.
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /io-control-spec-list/.
    
    ALLOCATE(read_stmt%io_control_spec_list)
    CALL ParseIOControlSpecList(  &
        tlist,  &
        3, closing - 1,  &
        .TRUE.,  &
        read_stmt%io_control_spec_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ! Continue on error.
    
    ! That can be it.
    IF (closing + 1 > SIZE(tlist)) THEN
      ALLOCATE(read_stmt%item_list(0))
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The /input-item-list/.
    
    CALL ParseIOItemList(  &
        tlist,  &
        closing + 1, SIZE(tlist),  &
        .FALSE.,  &           ! specify input-item-list.
        read_stmt%item_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! Check some constraints
    
    IF (ALLOCATED(read_stmt%io_control_spec_list%nml)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          LOCATION=QueryLocation(tlist(1)),  &
          COMPONENT='C916 on R913 via ' // comp,  &
          MSG='A /namelist-group-name/ shall not appear if the /read-stmt/ &
            &contains an /input-item-list/.' )
    END IF
    
  END SUBROUTINE parse_first_form
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the second form of /read-stmt/.
  !!
  !! This is the form that is the analogue of a /print-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    read_stmt         The results of parsing the /read-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_second_form(tlist, read_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE Formats
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ReadStmt), INTENT(OUT) :: read_stmt
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: comma          ! Index of the first comma.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ! Work out where the /format/ runs to.
    comma = FindBare(tlist, ittComma, scck_',', 2)
    IF (comma == 0) THEN
      it = SIZE(tlist)
    ELSE
      it = comma - 1
    END IF
    
    ALLOCATE(read_stmt%fmt)
    CALL ParseFormat(  &
        tlist,  &
        2, it,  &
        read_stmt%fmt,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    IF (comma == 0) THEN
      ALLOCATE(read_stmt%item_list(0))
      RETURN
    END IF
    
    it = comma + 1      ! Index of token after the comma.
    
    !---------------------------------------------------------------------------
    ! The /input-item-list/.
    
    CALL ParseIOItemList(  &
        tlist,  &
        it, SIZE(tlist),  &
        .FALSE.,  &           ! Specify input item list.
        read_stmt%item_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_second_form
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReadStmt%Dump - write a debugging representation 
  !! of the results of parsing a read statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE read_Dump(pnode, unit, pfx)
    
    USE Formats
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReadStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%io_control_spec_list)) THEN
      ! First form.
      WRITE (unit, "(A,'/io-control-spec-list/')") pfx
      CALL DumpIoControlSpecList( unit,  &
          pnode%io_control_spec_list, LEN(pfx) + 2 )
    ELSE
      ! Second form.
      WRITE (unit, "(A,'/format/ = ')", ADVANCE='NO')
      CALL pnode%fmt%Write(unit)
      WRITE (unit, "()")
    END IF
    
    WRITE (unit, "(A,'/input-item-list/')") pfx
    CALL DumpIOItemList(pnode%item_list, unit, pfx // '  ')
    
  END SUBROUTINE read_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE read_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReadStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%io_control_spec_list)) THEN
      CALL pnode%io_control_spec_list%Query(visitor)
    ELSE
      CALL pnode%fmt%Query(visitor)
    END IF
    
    CALL QueryIOItemList(pnode%item_list, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE read_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE read_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReadStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%io_control_spec_list)) THEN
      CALL pnode%io_control_spec_list%Modify(visitor)
    ELSE
      CALL pnode%fmt%Modify(visitor)
    END IF
    
    CALL ModifyIOItemList(pnode%item_list, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE read_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReadStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'read-stmt'.
  
  FUNCTION read_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'read-stmt'
    
  END FUNCTION read_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReadStmt%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE read_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReadStmt), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Component identifier for the /io-control-spec/.
    CHARACTER(*), PARAMETER :: ics_comp = 'R913 via ' // comp
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! These checks were originally part of IoControlSpec parsing, but 
    ! they were split out as part of splitting out checks from basic 
    ! parsing.  Unfortunately, after parsing is complex IoControlSpec has 
    ! no idea that it is for a read statement (perhaps it should), so these 
    ! checks were moved into the parent statement.
    
    IF (ALLOCATED(pnode%io_control_spec_list)) THEN
      ! @todo the location assigned to the following could be improved.
      IF (ALLOCATED(pnode%io_control_spec_list%delim)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C913 on ' // ics_comp,  &
            LOCATION=.LocationOf. pnode%io_control_spec_list%delim,  &
            MSG='A "DELIM=" specifier shall not appear in a /read-stmt/.' )
      END IF
    
      ! @todo the location assigned to the following could be improved.
      IF (ALLOCATED(pnode%io_control_spec_list%sign)) THEN
        CALL Add( err_list, CODE=errSyntax,  &
            COMPONENT='C913 on ' // ics_comp,  &
            LOCATION=.Locationof. pnode%io_control_spec_list%sign,  &
            MSG='A "SIGN=" specifier shall not appear in a /read-stmt/.' )
      END IF
    END IF
    
  END SUBROUTINE read_Check
  
END MODULE ReadStmts
