! $Id: ReturnStmts.f90 2161 2016-06-01 23:18:33Z ian $
!> @file
!! Defines the ReturnStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /return-stmt/'s.

MODULE ReturnStmts
  
  USE Expressions
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseReturnStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /return-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ReturnStmt
    !> The expression, if any, provided after the return keyword that 
    !! nominates the relevant alternate return.  Not allocated if there 
    !! was no such expression.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => return_Dump
    PROCEDURE :: Query => return_Query
    PROCEDURE :: Modify => return_Modify
    PROCEDURE, NOPASS :: What => return_What
  END TYPE ReturnStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of token that is a /return-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    expr              Optional integer expression on the 
  !! statement (indicates the alternate return).
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseReturnStmt(tlist, return_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
  
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ReturnStmt), INTENT(OUT) :: return_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------    
    ! Local constants
    
    ! Syntax rule number for a /return-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1241'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /return-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'RETURN', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (SIZE(tlist) == 1) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ParseExpr(tlist, 2, SIZE(tlist), return_stmt%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseReturnStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReturnStmt%Dump - write a debugging representation 
  !! of the results of parsing a return statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE return_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReturnStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%expr)) THEN
      WRITE (unit, "(A,'/scalar-int-expr/ = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%expr)
      WRITE (unit, "()")
    END IF  
    
  END SUBROUTINE return_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReturnStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE return_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReturnStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) THEN
      CALL pnode%expr%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE return_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReturnStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE return_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ReturnStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) THEN
      CALL pnode%expr%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE return_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ReturnStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'return-stmt'.
  
  FUNCTION return_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'return-stmt'
    
  END FUNCTION return_What
  
END MODULE ReturnStmts
