! $Id: SelectTypeConstructs.f90 2193 2016-06-06 13:01:31Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SelectTypeConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /select-type-construct/'s.

MODULE SelectTypeConstructs
  
  USE DeclarationTypeSpecs
  USE Expressions
  USE Identifiers
  USE Labels
  USE ParseNodes
  USE SelectTypeStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSelectTypeConstruct
  
  !-----------------------------------------------------------------------------
  
  !> Represents the results of parsing a case statement and its associated 
  !! block
  !!
  !! @todo Extend BlockStorage?
  TYPE :: case_stmt_block
    !> True if the type spec is a polymorphic specifier.  This will always 
    !! be true if the type_spec component is not allocated 
    !! (i.e. for CLASS DEFAULT).
    LOGICAL :: polymorphic
    !> The type spec in the type guard statement.
    !!
    !! Not allocated if this is the CLASS DEFAULT block.
    TYPE(TypeSpec), ALLOCATABLE :: type_spec
    !> The results of parsing the executable children.
    !!
    !! Always allocated, may be zero sized.
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
  END TYPE case_stmt_block
  
  !> Represents the results of parsing a /select-type-construct/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: SelectTypeConstruct
    !> The construct name, if present.
    !!
    !! Not allocated if not present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The associate name, if present.
    !!
    !! Not allocated if not present.
    TYPE(Identifier), ALLOCATABLE :: associate_name
    !> The selector.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: selector
    !> The type guard statements and following executable statements.
    !!
    !! Always allocated.  May be size zero if the user is being silly.
    TYPE(case_stmt_block), ALLOCATABLE :: blocks(:)
    !> The label on the end select type statement, if present, not allocated 
    !! otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => st_Dump
    PROCEDURE, NOPASS :: What => st_What
    PROCEDURE :: Query => st_Query
    PROCEDURE :: Modify => st_Modify
    PROCEDURE :: Advertise => st_Advertise
  END TYPE SelectTypeConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for a /select-type-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    construct         The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSelectTypeConstruct(st, construct, err_list)
    
    USE SelectTypeConstructStNodes
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SelectTypeConstructStNode), INTENT(INOUT) :: st
    TYPE(SelectTypeConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! case block index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening case statement.
    
    CALL ParseSelectTypeStmt(  &
        st%first%tlist,  &
        construct%associate_name,  &
        construct%selector,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Case blocks.
    
    ALLOCATE(construct%blocks(SIZE(st%blocks)))
    DO i = 1, SIZE(st%blocks)
      ASSOCIATE(child_st => st%blocks(i), child_pn => construct%blocks(i))
        CALL ParseTypeGuardStmt(  &
            child_st%initial%tlist,  &
            construct%construct_name,  &
            child_pn%polymorphic,  &
            child_pn%type_spec, sub_err_list )
        CALL Add(err_list, sub_err_list)
      
        CALL ParseBlock(child_st%exec, child_pn%children, sub_err_list)
        CALL Add(err_list, sub_err_list)
      END ASSOCIATE
    END DO
    
    !---------------------------------------------------------------------------
    ! The closing end select case statement.
    
    CALL ParseEndSelectTypeStmt(  &
        st%last%tlist,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      construct%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseSelectTypeConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of SelectTypeConstruct%Dump - write a debugging 
  !! representation of the results of parsing a select type construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE st_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SelectTypeConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! case block index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    IF (ALLOCATED(pnode%associate_name))  &
        WRITE (unit, "(A,'associate-name = ',A)") pfx,  &
          .NameOf. pnode%associate_name
    
    ! The selector.
    WRITE (unit, "(A,'selector = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%selector)
    WRITE (unit, "()")
    
    DO i = 1, SIZE(pnode%blocks)
      IF (ALLOCATED(pnode%blocks(i)%type_spec)) THEN
        WRITE (unit, "(A,'polymorphic = ',L1)") pfx,  &
            pnode%blocks(i)%polymorphic
        
        WRITE (unit, "(A,'type_spec = ')", ADVANCE='NO') pfx
        CALL WriteTypeSpec(unit, pnode%blocks(i)%type_spec)
        WRITE (unit, "()")
      END IF
      
      WRITE (unit, "(A,'child executable statements:')") pfx
      CALL DumpBlock(pnode%blocks(i)%children, unit, pfx // '  ')
    END DO
    
  END SUBROUTINE st_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SelectTypeConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'select-type-construct'.
  
  FUNCTION st_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'select-type-construct'
    
  END FUNCTION st_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SelectTypeConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE st_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SelectTypeConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%selector%Query(visitor)
    
    DO i = 1, SIZE(pnode%blocks)
      IF (ALLOCATED(pnode%blocks(i)%type_spec)) THEN
        CALL pnode%blocks(i)%type_spec%Query(visitor)
      END IF
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Query(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE st_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SelectTypeConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE st_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SelectTypeConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%selector%Modify(visitor)
    
    DO i = 1, SIZE(pnode%blocks)
      IF (ALLOCATED(pnode%blocks(i)%type_spec)) THEN
        CALL pnode%blocks(i)%type_spec%Modify(visitor)
      END IF
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Modify(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE st_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SelectTypeConstruct%Advertise - advertise the 
  !! identifiers referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE st_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SelectTypeConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE st_Advertise
  
END MODULE SelectTypeConstructs
