! $Id: StmtFunctionStmts.f90 2886 2019-04-06 22:41:52Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the StmtFunctionStmts module.


!*******************************************************************************
!!
!> Procedures for parsing statement function statements 
!! (/stmt-function-stmt/'s).
!!
!! This module also contains the standard statement parsing routine.

MODULE StmtFunctionStmts
  
  USE CompilerKinds
  USE Expressions
  USE Identifiers
  USE ParseNodes
  USE StmtFunctionStmtsUtils
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: IsMaybeStmtFunction   ! From StmtFunctionStmtsUtils.
  
  PUBLIC :: ParseStmtFunctionStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /stmt-function-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: StmtFunctionStmt
    !> Function name.
    TYPE(Identifier) :: function_name
    !> Dummy argument names.
    !!
    !! Always allocated but may be zero length.
    TYPE(Identifier), ALLOCATABLE :: args(:)
    !> The expression that defines the statement.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => stmtfn_Dump
    PROCEDURE :: Query => stmtfn_Query
    PROCEDURE :: Modify => stmtfn_Modify
    PROCEDURE, NOPASS :: What => stmtfn_What
  END TYPE StmtFunctionStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /stmt-function-stmt/ (R1243).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    stmt_fn           Results of parsing the statement 
  !! function.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseStmtFunctionStmt(tlist, stmt_fn, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(StmtFunctionStmt), INTENT(OUT) :: stmt_fn
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /stmt-function-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1243'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: closing      ! Index of the closing parentheses.
    INTEGER :: it           ! Current token index.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /stmt-function-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, 1, '/function-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    stmt_fn%function_name = IdentifierIfortbug(tlist(1))
    
    !---------------------------------------------------------------------------
    ! Arguments in parentheses
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing > 3) THEN
      CALL ParseNameList(  &
          tlist,  &
          3, closing - 1,  &
          '/dummy-arg-name/',  &
          stmt_fn%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ALLOCATE(stmt_fn%args(0))
    END IF
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! The "assignment" token.
    
    CALL Expect(tlist, it, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The expression.
    
    CALL ParseExpr(tlist, it + 1, SIZE(tlist), stmt_fn%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
  END SUBROUTINE ParseStmtFunctionStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of StmtFunctionStmt%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE stmtfn_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StmtFunctionStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Argument index.
   
    !***************************************************************************
    
    WRITE (unit, "(A,'/function-name/ = ',A)") .NameOf. pnode%function_name
    WRITE (unit, "(A,'args:')") pfx
    DO i = 1, SIZE(pnode%args)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%args(i)
    END DO
    WRITE (unit, "(A,'/expr/ = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, '()')
    
  END SUBROUTINE stmtfn_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of StmtFunctionStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE stmtfn_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StmtFunctionStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE stmtfn_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of StmtFunctionStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE stmtfn_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StmtFunctionStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE stmtfn_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of StmtFunctionStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'continue-stmt'.
  
  FUNCTION stmtfn_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'stmt-function-stmt'
    
  END FUNCTION stmtfn_What
  
END MODULE StmtFunctionStmts
