! $Id: TargetStmts.f90 2881 2019-04-06 00:12:06Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the TargetStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /target-stmt/'s (R556).

MODULE TargetStmts
  
  USE Identifiers
  USE SubscriptLists
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseTargetStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /target-decl/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: TargetDecl
    !> The name of the entity.
    TYPE(Identifier) :: name
    !> The /array-spec/ - not allocated if not present.
    TYPE(DimSpec), ALLOCATABLE :: array_spec(:)
    !> The /coarray-spec/ - not allocated if not present.
    TYPE(DimSpec), ALLOCATABLE :: coarray_spec(:)
  CONTAINS
    PROCEDURE :: Write => td_Write
    PROCEDURE :: Query => td_Query
    PROCEDURE :: Modify => td_Modify
    PROCEDURE, NOPASS :: What => td_What
    PROCEDURE :: Advertise => td_Advertise
    PROCEDURE :: GetLocation => td_GetLocation
  END TYPE TargetDecl
  
  !> Stores the results of parsing a /target-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: TargetStmt
    !> The list of entities being given the target attribute.
    TYPE(TargetDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => targ_Dump
    PROCEDURE :: Query => target_Query
    PROCEDURE :: Modify => target_Modify
    PROCEDURE, NOPASS :: What => targ_What
  END TYPE TargetStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /target-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    targ_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTargetStmt(tlist, targ_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !--------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TargetStmt), INTENT(OUT) :: targ_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /target-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R556'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /target-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /target-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! TARGET...
    
    CALL Expect(tlist, 1, ittName, scck_'TARGET', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /target-decl/.  A /target-decl/ starts with an /object-name/.
    
    CALL Expect(  &
        tlist,  &
        2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /object-name/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /target-decl-list/ here.
    
    !---------------------------------------------------------------------------
    ! /target-decl-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(targ_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_target_decl(  &
          tlist,  &
          item_start, item_finish,  &
          targ_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParseTargetStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetStmt%Dump - write a debugging representation 
  !! of the results of parsing a target statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE targ_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Decl index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL pnode%decls(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE targ_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE target_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE target_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE target_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE target_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'target-stmt'.
  
  FUNCTION targ_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'target-stmt'
    
  END FUNCTION targ_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for TargetDecl.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /target-decl/ (R557).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    decl              Results of parsing the /target-decl/.  
  !! Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Error messages assume that this is being called as part of processing a 
  !! comma separated list.
  
  SUBROUTINE parse_target_decl(tlist, start, finish, decl, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(TargetDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /target-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R557'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: it             ! Current token index.
    
    ! Error list for ParseDeferredShapeSpecList call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! /object-name/ ...
    
    CALL Expect(tlist, start, '/object-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! Either that is it, or we have ( /array-spec/), or [ /coarray-spec/ ]
    
    CALL Expect(  &
        tlist(:finish),  &
        start + 1,  &
        [ittEndStatement, ittDelimiter, ittDelimiter],  &
        [scck_' ', scck_'(', scck_'['],  &
        'end-of-statement, ",", "(" or "["',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) RETURN    ! That's it.
    
    !---------------------------------------------------------------------------
    ! ( /array-spec/ )
    
    IF (opt == 2) THEN
      CALL ExpectClosing(tlist, start + 1, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL ParseArraySpec(  &
          tlist(start+2:closing-1), .FALSE.,  &
          decl%array_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      ! Either that is it, or we have a [ /coarray-spec/ ].
      CALL Expect(  &
          tlist(:finish),  &
          closing + 1,  &
          [ittEndStatement, ittDelimiter],  &
          [scck_' ', scck_'['],  &
          'end-of-statement, ",", or "["',  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (opt == 1) RETURN      ! That's it.
      
      it = closing + 1
    ELSE
      it = start + 1
    END IF
    ! it here the opening bracket of the coarray-spec.
    
    !---------------------------------------------------------------------------
    ! [ /coarray-spec/ ]
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseCoarraySpec(  &
        tlist(it+1:closing-1),  &
        decl%coarray_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Check nothing superfluous after the closing "]".
    
    CALL ExpectEOS(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_target_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE td_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('/object-name/ = ',A)",  ADVANCE='NO') &
          .NameOf. pnode%name
    IF (ALLOCATED(pnode%array_spec)) THEN
      WRITE (unit, "(' (')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%array_spec)
      WRITE (unit, "(')')", ADVANCE='NO')
    END IF
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      WRITE (unit, "(' [')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%coarray_spec)
      WRITE (unit, "(']')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE td_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE td_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! [Co]array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE td_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE td_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! [co]array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE td_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'target-decl'.
  
  FUNCTION td_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'target-decl'
    
  END FUNCTION td_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION td_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(TargetDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION td_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of TargetDecl%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE td_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TargetDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    IF (ALLOCATED(pnode%array_spec) .AND. ALLOCATED(pnode%coarray_spec)) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[  &
            NatureDefTarget,  &
            NatureDefDimension,  &
            NatureDefCodimension ], &
          LOCATIONS=[  &
            .LocationOf. pnode%name,  &
            .LocationOf. pnode%name,  &
            .LocationOf. pnode%name ] )
    ELSE IF (ALLOCATED(pnode%array_spec)) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefTarget, NatureDefDimension],  &
          LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    ELSE IF (ALLOCATED(pnode%coarray_spec)) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefTarget, NatureDefCodimension],  &
          LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefTarget], &
          LOCATIONS=[.LocationOf. pnode%name] )
    END IF
    
  END SUBROUTINE td_Advertise
  
END MODULE TargetStmts
