! $Id: ValueStmts.f90 2196 2016-06-07 02:56:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ValueStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /value-stmt/'s.

MODULE ValueStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseValueStmt
  
  !> Stores the results of parsing a /volatile-stmt/.
  TYPE, EXTENDS(NameListPNode), PUBLIC :: ValueStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => val_GetListName
    PROCEDURE, NOPASS :: What => val_What
    PROCEDURE :: Advertise => val_Advertise
  END TYPE ValueStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /value-stmt/ (R558).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    val_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseValueStmt(tlist, val_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ValueStmt), INTENT(OUT) :: val_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /optional-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R558'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /value-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! OPTIONAL...
    
    CALL Expect(tlist, 1, ittName, scck_'VALUE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /dummy-arg-name/
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /dummy-arg-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /dummy-arg-name-list/ here.
    
    !---------------------------------------------------------------------------
    ! The /dummy-arg-name-list.
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/dummy-arg-name/',  &
        val_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseValueStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ValueStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/dummy-arg-name-list/'
  
  FUNCTION val_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/dummy-arg-name-list/'
    
  END FUNCTION val_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of ValueStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'value-stmt'.
  
  FUNCTION val_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'value-stmt'
    
  END FUNCTION val_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ValueStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE val_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ValueStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[NatureDefValue], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE val_Advertise
  
END MODULE ValueStmts
