! $Id: WhereStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the WhereStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /where-stmt/'s (R741), 
!! /where-construct-stmt/'s, /masked-elsewhere-stmt/'s, /elsewhere-stmt/'s 
!! and /end-where-stmt/'s.
!!
!! This module has similar tasks to the IfStmts module.

MODULE WhereStmts
  
  USE AssignmentStmts
  USE Expressions
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseWhereStmt
  PUBLIC :: ParseWhereConstructStmt
  PUBLIC :: ParseMaskedElsewhereStmt
  PUBLIC :: ParseElsewhereStmt
  PUBLIC :: ParseEndWhereStmt
  
  PUBLIC :: Expression        ! From Expressions module for convenience.
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /where-stmt/ (R741).  This is the 
  !! single line form of where statement.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /if-stmt/.
  !!
  !! @param[out]    mask_expr         The expression that is tested by the 
  !! /where-stmt/.
  !!
  !! @param[out]    child_tlist       List of tokens that make up the child 
  !! /action-stmt/ of the /where-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWhereStmt(tlist, mask_expr, child_tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: mask_expr
    TYPE(Token), INTENT(OUT), ALLOCATABLE :: child_tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /where-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R741'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /where-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! WHERE ....
    
    CALL Expect(tlist, 1, ittName, scck_'WHERE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! WHERE(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The mask expression.
    
    CALL ParseExpr(tlist, 3, closing - 1, mask_expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! The child assignment statement.
    
    IF (SIZE(tlist) == closing) THEN
      ! The error message is cognisant of classification options.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(closing)),  &
          MSG='Expected /where-assignment-stmt/ after WHERE (...).' )
      RETURN
    END IF
    
    child_tlist = tlist(closing+1:)
    
  END SUBROUTINE ParseWhereStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /where-construct-stmt/ (R741).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /where-stmt/.
  !!
  !! @param[out]    mask_expr         The expression that is tested by the 
  !! /where-stmt/.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /where-construct-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWhereConstructStmt( tlist, mask_expr, construct_name,  &
        err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: mask_expr
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /where-construct-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R743'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /where-construct-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! WHERE ....
    
    CALL ExpectKeyword(tlist, it, scck_'WHERE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! WHERE(...)
    
    CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The mask expression.
    
    CALL ParseExpr(tlist, it + 2, closing - 1, mask_expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Thats it!
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseWhereConstructStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /masked-elsewhere-stmt/ (R747).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /masked-elsewhere-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the 
  !! opening /where-construct-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    mask_expr         The expression that is tested by the 
  !! /masked-elsewhere-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseMaskedElsewhereStmt(tlist, construct_name, mask_expr, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: mask_expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /masked-elsewhere-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R747'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /masked-elsewhere-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ELSEWHERE or ELSE WHERE
    
    CALL Expect(tlist, 1, scck_'ELSE', scck_'WHERE', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... ( expr )
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, it + 1, closing - 1, mask_expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    !---------------------------------------------------------------------------
    ! ... construct-name
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        closing + 1,  &
        construct_name,  &
        '/where-construct-stmt/',  &
        '/masked-elsewhere-stmt', &
        'C732 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseMaskedElsewhereStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /elsewhere-stmt/ (R748).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /elsewhere-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the 
  !! opening /where-construct-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseElseWhereStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /elsewhere-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R748'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /elsewhere-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ELSEWHERE or ELSE WHERE
    
    CALL Expect(tlist, 1, scck_'ELSE', scck_'WHERE', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional construct name
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/where-construct-stmt/', '/elsewhere-stmt', &
        'C732 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseElseWhereStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-where-stmt/ (R749).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-where-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /where-construct-stmt/ if one was present, not allocated if a construct 
  !! name was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndWhereStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-where-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R749'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-where-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDWHERE or END WHERE ...
    
    CALL Expect(tlist, 1, scck_'END', scck_'WHERE', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/where-construct-stmt/',  &
        '/end-where-stmt/',  &
        'C732 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndWhereStmt
  
END MODULE WhereStmts
